/**     @file iocount.h
	MASS project: header file for iocount.cpp
 
	This module implements the Input/Output operations with disk for saving and
	restoring the counts in count buffers of channels. The aim is to allow for
	the saving the night-time count series in a binary file and for the \em 
	off-line playback the recorded series of input channel counts to repeat
	their processing and displaying the output parameters.

	The counts from all channels of the device are written in binary format by
	the utility ioc::write() which returns the \em start-position \em address
	of the written series. This address is saved somewhere else (e.g. in main
	MASS output file). Then, upon reading that file, the address is fed to
	ioc::read() which restores the counts from the file in the channel buffers.
	After this, the processing may go on as if the count data were just
	received from counters.
	
	The binary count file should be opened with an attribute "a+b" and closed
	upon completion of reading/writing externally. On error (bad input address
	for ioc::read() or low-level I/O problem), the error is set by utilities to
	be detected by nr::erget() and nr::ermessage().

	The buffers are made of the type \em count_t which should be specified
	externally. All the bytes of count numbers are written to the disk. Thus,
	the two-byte unsigned or signed integers are seemed to be best suited for
	MASS, for which we do in this module:
	\verbatim
		#include <stdint.h> // or <sys/types.h>
		...
		typedef count_t	uint16_t ;	// or int16_t
	\endverbatim
	
	@author	N.Shatsky (kolja@sai.msu.ru)
	@version	1.0
*/
#ifndef _IOCOUNT_H
#define _IOCOUNT_H

/** Include the definitions of the fixed-length integers: may alternatively be
<sys/types.h>
*/
#include <stdint.h>

#include <stdio.h>

/** The count type will be unsigned 16-bit integer type */
typedef uint16_t count_t ;
/** Maximal count_t (implementation works for unsigned integral types) */
const count_t max_count_t = (count_t)(-1) ; 

/**	@brief
	Declarations of the channel count type and counts I/O operations */
namespace ioc {

/**
	@brief	Write the channel count series to disk
	@param	fcnt	structure of the opened binary file of counts
	@param	buffer	array of \em nbuf pointers to channel count buffers
	@param	buflen	length of the channel buffer in units of counts (not bytes!)
	@param 	nbuf	Number of buffers (e.g. number of active channels)
	@return	Starting position in file (in bytes from beginning) where
		the counts were written or (-1) on error
	
	The function uses the \em fseek() to set the file position to the end of
	file (where it is, normally, set already), gets this current position to be
	returned on success, and writes sequentially all \em nbuf channel buffers to
	disk. 
*/
long write(FILE * fcnt, count_t ** buffer, size_t buflen, int nbuf) ;

/**
	@brief	Read the channel count series from disk
	@param	fcnt	structure of the opened binary file of counts
	@param	buffer	array of \em nbuf pointers to channel count buffers
	@param	buflen	length of the channel buffer in units of counts (not bytes!)
	@param 	nbuf	Number of buffers (e.g. number of active channels)
	@param	addr	Starting position in file (in bytes from beginning) where
		the counts should be read from. 
		
	The address is checked to be only the multiple of the size of the \em
	count_t type. All the rest possible non-synchronizations are not tested. The
	error is possible when not all the counts are successfully read.
	
	The function uses the \em fseek() to set the file position to the given \em
	addr, and reads sequentially all \em nbuf channel buffers from disk into
	the respective \em buffer[iCounter].
*/
void read(FILE * fcnt, count_t ** buffer, size_t buflen, int nbuf, long addr) ;

}	// end of namespace ioc

#endif
