/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            atmos.c                                       */
/*                    Description is given in atmos.h                       */
/* Editor tabular size: 4                                                   */
/****************************************************************************/
 
/* Insert the definition ATMOSTEST in compilation for the module self-tests:
gcc -c atmos.c .... -DATMOSTEST=1 */
//#define ATMOSTEST

#include <math.h>
#include "atmos.hxx"
#include "weif.hxx"
#include "nrutil.h"
#include "nr.h"

#ifdef ATMOSTEST
#define NTEST 10

/* Display the intermediate information for checking the define-settings */
/** Print names of shifted weights in use, number of base 
functions after small singulars rejection etc. */
//#define PRNBASEINFO	
/** Print the coefficients of decomposition by orthonormal basis: */
//#define PRNCOEF
/** Print the plot data for "atmos | xmgrace -pipe" for examination of
PRNAPPROX-th moment approximation: */
//#define PRNAPPROX 3

#if defined PRNAPPROX || defined PRNCOEF || defined PRNBASEINFO
	#undef NTEST
	#define NTEST 1
#endif
/*---------------------------- TEST PART ---------------------------------*/

/*----------- Comes normally from configuration file: cf_... vars ----------*/

/* Apertures: */
/* number of apertures in segmentor : */
int cf_naper = 4;
/* Physical size of segmentor annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.102, 0.188, 0.354};
double cf_aper_out[] = { 0.100, 0.185, 0.350, 0.650};
//double cf_aper_inn[] = { 0.000, 0.100};
//double cf_aper_out[] = { 0.100, 0.200};
/* Magnification of the system: */
double cf_magnif = 20.0 ;

/* Proportional Altitude grid: */
double cf_z0 = 0, cf_zmax = 25.0, cf_dz = 0.2, cf_dzmin = 0.2;
/* Equidistant Altitude grid: */
//double cf_z0 = 0, cf_zmax = 25.0, cf_dz = 0.5, cf_dzmin = 0.0;

/* Alternative grid: to interpolate to */
double int_z0 = 0, int_zmax = 20.0, int_dz = 0.02, int_dzmin = 0;

double cf_shift = -0.5 ;	/* Shift weight functions by cf_shift kilometers */

const int NIDX = 10 ;
const int NSAMIDX = NIDX*2+1 ;
double samidx[NSAMIDX] = {
#include "a05_idx.dat"	/* uncomment one line in this file */
,
#include "a05_sidx.dat"	/* uncomment one line in this file */
, 0.001	/* shifted desi */
};
const double IDXRELERR = 0.034 ;	/* Relative error of instantaneous indices */

const int CF_NMEAS = 10 ;
char * fw = "weight.new" ;

char fcrv[] = "mass.crv" ;	/* System response curve */
char fsed[] = "../weif/spectra/a05.sp" ;	/* Energy distrib. file */
//char fsed[] = "a05.sed" ;	/* Energy distrib. file */

int main( int ac, char **av ) {

int err, idum = (-14) ;
double idx[NSAMIDX], aidx[NSAMIDX], e2idx[NSAMIDX] ;
const char foutname[] = "atmos.out" ;

FILE * fout = fopen(foutname,"at");
	
if (!fout) {
	fprintf(stderr,"%s open error\n",foutname);
	return -1 ;
}

atm::init(
	cf_naper, cf_magnif, cf_aper_inn, cf_aper_out,
	cf_z0, cf_zmax, cf_dz, cf_dzmin,
	fcrv) ;

/* write headings: */
atm::write(fout,"#A", atm::ATMPAR, atm::HEAD) ;
atm::write(fout,"#T", atm::CN2PROF, atm::HEAD) ;


for (int test=0; test<NTEST; test++) {
	err = atm::loadsed(fsed) ;
	if (err<0) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str());
		return err ;
	}
	
	err = atm::checkwf(fw) ;
	if (err) {
		nr::erreset() ;
		printf("Start calcwf()\n");
		atm::calcwf(fw, wf::progress) ;
	}

	atm::update(fw, "tmp.txt", 0, cf_shift) ;

	/* One check per all calls: */
	if (nr::erget()) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
		return nr::erget() ;
	}

	for (int i=0; i<NSAMIDX; i++) aidx[i] = e2idx[i] = 0 ;
	
	atm::alloc(CF_NMEAS) ;
	
	int meas = 0 ;	/* we need this in avgint() after loop */
	for (meas=0; meas<CF_NMEAS; meas++) {
		/* Emulate the noisy indices: */
		for (int i=0; i<NSAMIDX; i++) {
			idx[i] = samidx[i]  * (1.0 + nr::gasdev(&idum) * IDXRELERR) ; 
			aidx[i] += idx[i] ;
			double d = (idx[i] - samidx[i]) ;
			e2idx[i] += d*d ;
		}

		atm::calcint(idx, NULL) ;
	}
	double avgreldev = 0 ;
	for (int i=0; i<NSAMIDX; i++) {
		e2idx[i]  /= (CF_NMEAS * (CF_NMEAS-1)) ;
		aidx[i] /= CF_NMEAS ;
		double reldev = (aidx[i]-samidx[i])*100/samidx[i] ;	// percents
		avgreldev += reldev*reldev ;
//		double relerr = sqrt(e2idx[i])*100/aidx[i] ;
//		printf("idx#%d: Expected=%7.4f: reldev(avg)=%5.2f%% relerr=%5.2f%%\n",
//			i,samidx[i],reldev,relerr);

		/* reset the error into the constant relative: */
		double d = IDXRELERR*aidx[i] ;
		e2idx[i] = d*d ;
	}
	avgreldev = sqrt(avgreldev/NSAMIDX) ;
//	printf("RMS rel.dev. among indices is %5.2f%%\n",avgreldev);
	for (int meth=0;meth<=1;meth++) {
		double chi2=atm::calcn2(aidx, &(aidx[NIDX]) ,e2idx,&(e2idx[NIDX]),meth) ;

		printf("#   Chi2");
		for (int h=0; h<atm::getncn2(); h++) printf("\tz_%d",h);			
		for (int h=0; h<atm::getncn2(); h++) printf("\tCn2_%d",h);			
		printf("\n");
		printf("T%d %5.2f\t",meth+1, chi2);
		for (int i=0; i<atm::getncn2(); i++)
			printf("%2d\t",(int)(atm::getzcn2(i)+0.5)) ;
		for (int i=0; i<atm::getncn2(); i++)
			printf("%7.1e\t",atm::getcn2(i));
		printf("\n");
	}
	atm::avgint(aidx[20],e2idx[20],0.0) ;

	/* Output the parameters as a single table: */
	printf("# fSee\tSee\tTheta_0\tfHeff\tHeff\tM_2*e5\tTconst\n");
	printf("A %5.2f\t%5.2f\t%5.2f\t%5.0f\t%5.0f\t",
		atm::getval(atm::FSEE), atm::getval(atm::WSEE), atm::getval(atm::ISP),
		atm::getval(atm::FHEFF), atm::getval(atm::HEFF));
		printf("%6.3f\t%5.2f\n", atm::getval(atm::M2)*1e5, atm::getval(atm::TC));
	atm::write(fout,"A ", atm::ATMPAR) ;
	atm::write(fout,"T ", atm::CN2PROF) ;

	/* One check per all calls: */
	if (nr::erget()) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
		return nr::erget() ;
	}

	printf("#========================= %d==========================\n",test);
}
atm::done() ;
	
fclose(fout) ;	
return 0 ;
}
#endif

/** private part of a namespace at */

namespace atm {

/*------------------------ Global parameters -----------------------------*/

/* Bad result placeholder */
double bad = 99 ;

/** Free atmosphere seeing computed by calcint() */
double fseeing = 0, efseeing = 0 ;

/** Whole atmosphere seeing computed by calcint() */
double seeing = 0 , eseeing = 0 ;

/** Isoplanatic angle  computed by calcint() */
double isoplan = 0 , eisoplan = 0 ;

/** Effective altitude of the turbulence  from calcint(): */
double heff = 0 , eheff = 0 ;

/** Effective altitude in free atmosphere from calcint(): */
double fheff = 0 , efheff = 0 ;

/** atmospheric time constant [ms]  from calcint() */
double timeconst = 0 , etimeconst = 0 ;

/** Length of the computed Cn^2-vector from calcn2() */
int ncn2 = 0 ;

/** Cn^2-vector computed by Cn2 restoration calcn2() and Chi2 as last elem.*/
double * cn2 = 0 ;

/** Altitudes of Cn^2 layers from calcn2() */
double * zcn2 = 0 ;

/** Basic set of weight functions: */
wf_t atwf ;

/** User-set altitude grid: */
//double z0 = wf::Z0, zmax = wf::ZMAX, dz=wf::DZ, dzmin=wf::DZMIN ; 
double z0 = 0, zmax = 0, dz=0, dzmin=0; 

/** Response function of the atmosphere + feeding optics + MASS system */
double * response = 0 ;

/* Currently loaded weight file name: */
string curwffile = "" ;

/** Response function wavelength grid */
double * lambda = 0 ;

/** lambda's and response' grid length: */
int nresponse = 0 ;

/** Current star spectrum multiplied by response curve response */
double * sed = 0 ;

/** Physical length of current star spectrum array */
int nsed = 0 ;

/** Two Matrices of coefficients for the moments of Cn2  */
double** coef[2] = {NULL, NULL} ;	/**< [0]: for non-shifted measurements, [1] 
for generalized mode measurements */

/** Number of photometric channels */
int nchan = 0 ;

/** Number of weight functions used in normal and generalized modes (the 
 column number of the matrix \em coef) */
int nweight[NMODE] = {0,0} ;

/** Shifts by altitude by which normal (first) and generalized (both) modes 
operate */
double zshift[NMODE] = {0,0} ;

/* Global combined matrix of weights for restoration of the profile: non-shifted
and shifted weights togather, with 1km spacing */
double ** wcomb = 0 ;

/* Global combined average indices and their errors for restoration of
Cn2-profile */
double * sicomb = 0, * e2sicomb = 0 ;

/* Matrix of individual integrals */
double ** cn2int = 0 ;

/* Number of allocated rows in int[][] */
int maxnmeas = -1 ;

/* current number of measurements of integral put in storages: */
int curmeas = -1 ;

/* Vectors of average integrals and their errors */
double * aint = 0 ;
double * e2int = 0 ;

/* File-reader for spectra and response curve */
void read2col(const char * filename, double ** col1,
	double ** col2, int * nrow) ;

void atm::conwgrid(double * wgrid, int n, int scale) ;

void atm::sedcrv(double * lambda_sed, double * sed, int nsed) ;	

}

/*============================== Functions ===============================*/

/*------------------------------ getval ----------------------------*/

double atm::getval(ordint what) {
	switch (what) {
		case FSEE:	return fseeing;
		case WSEE:	return seeing ;
		case ISP:	return isoplan ;
		case M2:	return aint[IISK] ;
		case FHEFF:	return fheff ;
		case HEFF:	return heff ;
		case TC:	return timeconst ;
		default:
			nr::nrerror("ATMOS: getval(): wrong what-selection",nr::ERPAR);
			return 0 ;
	}
}

/*------------------------------ geterr ----------------------------*/

double atm::geterr(ordint what) {
	switch (what) {
		case FSEE:	return efseeing;
		case WSEE:	return eseeing ;
		case ISP:	return eisoplan ;
		case M2:	return e2int[IISK]>0 ? sqrt(e2int[IISK]) : 0;
		case FHEFF:	return efheff ;
		case HEFF:	return eheff ;
		case TC:	return etimeconst ;
		default:
			nr::nrerror("ATMOS: geterr(): wrong what-selection",nr::ERPAR);
			return 0 ;
	}
}

/*------------------------------ getncn2 -------------------------------*/

int atm::getncn2() {
	return ncn2 ;
}

/*------------------------------ getzcn2 -------------------------------*/

double atm::getzcn2(int i) {
	if (zcn2)
		return zcn2[i] ;
	else {
		nr::nrerror("ATMOS: getzcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ getcn2 --------------------------------*/

double atm::getcn2(int i) {
	if (cn2)
		return cn2[i] ;
	else {
		nr::nrerror("ATMOS: getcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ geter2int ------------------------------*/

double atm::geter2int(int i) {
	if (e2int)
		return e2int[i]>0 ? e2int[i] : 0 ;
	else {
		nr::nrerror("ATMOS: geter2int() while uninitialized",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ read2col --------------------------------*/

/**
	@brief	Read the two columns of argument and function values from file
	@param	filename	file name
	@param	col1	pointer to the array of arguments (first column data)
	@param	col2	pointer to the array of function values (second column)
	@param	nrow	pointer to the place where to put the length of allocated 
		arrays of arguments and function values
	
	This utility reads the pairs of numeric values from the text-type file
	ignoring the lines beginning with comment symbol "#". Two non-allocated
	pointers must be supplied as NULLs on start. Function counts the non-comment
	data lines, allocates the memory for arrays of argument (first column) and
	function values (second column) with lengths equal to the number of the 
	counted data lines, and then reads the columns into them.
	
	The following checks are done on each non-comment line:
	- at least two numeric tokens are given in each line (argument and function)
	- argument is a monotonously rising function
	- argument values start from some positive number
	- function is non-negative 
	- function has a positive average
	The function is thus suited for reading the spectrum-like information.
	
	Errors possible are the problem of file opening/reading or line parsing. In
	case of error, the pointers are returned unallocated.
*/	
void atm::read2col(const char * filename, double ** col1,
	double ** col2, int * nrow) {

	const int DATALINE = 100 ;      /* longest feasible line */

	if (nr::erget()) return ;

	/* Protect from allocation on allocated pointer */
	if ((*col1!=NULL) || (*col2!=NULL)) {
		nr::nrerror("ATMOS: read2col() on non-NULL storages", nr::ERNNL) ;
		return ;
	}

	if (!filename) {
		nr::nrerror("ATMOS: read2col() with no file name", nr::ERPAR) ;
		return ;
	}
	
	/* Count col1 steps: */
	FILE * f = fopen(filename,"rt") ;
	string msg = "ATMOS: read2col() with file " ;
	msg += filename ;
	if (!f)	{
OPENERR:
		msg += ": open error" ;
		nr::nrerror(msg.c_str(), nr::ERFIO) ;
		return ;
	}
		
	int ndata = 0;
	char buff[DATALINE];
	while (fgets(buff,DATALINE,f)) {
		char * id = strtok(buff, wf::FDELIM) ; 
		if (id && (id[0]!='#')) 
			/* it is not a comment and there is at least one item */
			ndata++;
	}
	fclose(f);
	
	if (!ndata) {
NODATA:
		msg += ": contains no data" ;
		nr::nrerror(msg.c_str(), nr::ERNOD) ;
		return ;
	};
	
	/* Reopen the file : */
	f = fopen(filename,"rt") ;
	if (!f) goto OPENERR;

	/* Allocate memory and, if fails, close the file and return: */
	(*col1)=nr::dvector(0,ndata-1) ;
	(*col2)=nr::dvector(0,ndata-1) ;
	(*nrow)=ndata ;

	int err = 0;
	int nline = 0 ;
	ndata = 0;
	double prevarg = 0, sum = 0 ;	/* previous argument and function \int storages */
	while (fgets(buff,DATALINE,f)) {
		nline++ ;
		char * id = strtok(buff, wf::FDELIM) ; 
		if (id && (id[0]!='#')) {
			/* it is not a comment : try to read data now: */

			/* argument: */
			char * endptr ;
			(*col1)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || !(id = strtok(NULL, wf::FDELIM)) || 
				((*col1)[ndata]<=prevarg)) {
				/* Argument token is not a number or 
				   previous argument is not less than current or
				   function value is absent: return line number */
				err = nline ;
				break ;
			}
			
			/* function: */
			(*col2)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || ((*col2)[ndata]<0)) {
				/* Function token is not a number or
				   function is negative */
				err = nline ;
				break ;
			}
			sum += (*col2)[ndata] ;
			prevarg = (*col1)[ndata] ;
			ndata++;
		}
	}
	fclose(f);
	/* Check that the function is positive (not zero): */
	if (!err && (sum<=0)) goto NODATA ;
	if (err) {
		/* Reading or conversion or argument/function constraints error: */
		nr::free_dvector(*col1,0,*nrow-1);
		nr::free_dvector(*col2,0,*nrow-1);
		(*col1) = (*col2) = 0 ;
		sprintf(buff,": error in line %d",err) ;
		msg += buff ;
		nr::nrerror(msg.c_str(), nr::ERSPE) ;
		return ;
	} 
}

/*-------------------------------- conwgrid -------------------------------*/

/**
	@brief	Convert array of optical (!) wavelengths into given scale
	@param	wgrid	array of optical wavelengths in [angstrem] or [nm]
	@param	n	its length
	@param	scale	decimal logarithm of the wavelength of the infrared end of 
		the optical range to obtain
	
	This is an utility for conversion of the spectral curves (response function
	or spectrum) into the wavelength grid in a given units \em scale. It takes
	the value from the middle of the array and assumes that it must be in
	between 0.1 and 1mkm if it's an optical range. Then it makes the conversion by
	multiplication of the array value by an integer power of 10 which makes the
	middle array value lying in the range (0.1--1.0)*10**{scale}.
	
	\b Example
	
	Let the middle point of \em wgrid (it is \em wgrid[n/2]) be equal to 410 on
	start (i.e. the grid is in [nm]). \em Scale is given as "0", i.e. we want
	the spectrum to belong to the range 0.1--1.0, i.e. to be expressed in [mkm]
	on exit. The initial range is \em range=(int)lg(410)+1=3. So, the
	conversion coefficient will be \em 10**conv, where \em conv=scale-range=-3.
*/
void atm::conwgrid(double * wgrid, int n, int scale) {

	if (nr::erget()) return ;

	if ((wgrid==NULL) || (n<1)) {
		nr::nrerror("ATMOS: conwgrid() on null data", nr::ERNUL) ;
		return ;
	}
	
	double conv = scale - ((int)log10(wgrid[n/2])+1) ;
/* printf("Power conversion = %3.0f\n",conv);*/
	conv = pow(10.0,conv);
/*printf("Conversion = %7.1e\n",conv);*/
	for (int i=0; i<n; i++) wgrid[i] *= conv;
}

/*------------------------------ init() --------------------------------*/

void atm::init(
	int nchan, double diam, double *eps_inn, double *eps_out,
	double uz0, double uzmax, double udz, double udzmin,
	const char * responsefile) {

const int MKM_RANGE	= 0	;
/**< Units of optical range wavelength: 0 for [mkm], 3 for [nm], 4 for [\AA] */

	if (nr::erget()) return ;

	if (atm::coef[0]) {
		/* done() was not called in advance! */
		done() ;
	}

	if ((eps_inn==NULL) || (eps_out==NULL)) {
		nr::nrerror("ATMOS: init() on null aperture sizes", nr::ERNUL) ;
		return ;
	}

	/* Read the response function from the file: */
	read2col(responsefile,&atm::lambda,&atm::response,&atm::nresponse) ;
	/* Convert whatsoever (nm or angstrems) into microns: */
	conwgrid(atm::lambda, atm::nresponse, MKM_RANGE) ;
	if (nr::erget()) return ;	/* returns if read2col() or conwgrid() fail */

	/* Set the altitude grid and the aperture geometry in basic wf */	
	atwf.setzgrid(uz0, uzmax, udz, udzmin) ;
	int nz = atwf.getnz() ;

	/* Assign user-supplied zgrid parameters for update(): */
	atm::z0 = uz0, atm::zmax = uzmax, atm::dz = udz; atm::dzmin = udzmin ;
	
	atwf.setaper(nchan, diam, eps_inn, eps_out) ;
	int nw = atwf.getnw() ;

	/* Compute the total number of weight functions that may be in use: */
	atm::nweight[0] = nw ; /* Normal mode: one z-shift */
	
	atm::nweight[1] = nw ; /* Generalized mode: two z-shifts */
	for (int i=0; (i<(int)strlen(atm::USESHFT)) && (i<nw); i++) 
		if (atm::USESHFT[i]!='0') (atm::nweight[1])++ ;

	/* Check that the total number of weights will be less than the number of
	altitude steps in orthogonal matrix: */
	if (atm::nweight[1]>=nz) {
		atm::done() ;
		nr::nrerror("ATMOS: too few steps in altitude",nr::ERZGR) ;
		return ;
	}

	/* Create the combined weight matrix for Cn2-restoration with 1km-cell grid.
	It works for both non-shifted mode (its first nw rows only) and for
	generalized mode (full matrix). */
	/* Number of z-steps in restoration grid: */
	nz = (int)(ceil((floor(zmax)-floor(z0))/DZCN2+1)) ;
	wcomb=nr::dmatrix(0,nweight[1]-1,0,nz-1) ;

	/* Creation of the matrices of decomposition coefficients */
//	coef = new (double**)[NMODE] ;
	for (int i=0; i<NMODE; i++)
		coef[i] = nr::dmatrix(0,NPOWER-1,0,nweight[i]-1) ;
		
	atm::nchan = nchan ;

}

/*---------------------------- sedcrv ---------------------------------*/
/**
	@brief	Multiply the spectrum curve by the loaded response curve with 
		optionally shifted origins
	@param	lambda_sed	array of wavelengths of spectrum
	@param	sed  	array of spectrum
	@param	nsed	length of \em sed[] and \em lambda_sed[]
	
	Function looks for the coincidence of some \em k-th element of \em 
	lambda_sed[] grid with the first element of the response curve wavelength
	grid \em lambda_crv. This \em k is a shift (or difference) of origins of
	SED and response curves (normally, the spectrum is wider). 

	Then, first elements of spectrum - \em sed[i], i=0..nresponse - are
	rewritten with the product of SED and response curves 
	\em	sed[i+k]*response[i] which takes into account the different origins of
	these functions. Last \em nsed-nresponse elements of \em sed[] are replaced with
	zeros. Thus, on exit only first \em nresponse elements of \em sed[] make
	sense.
	
	@see calcwf() checksed()
*/
void atm::sedcrv(double * lambda_sed, double * sed, int nsed) {

	if (nr::erget()) return ;

	if ((lambda_sed==NULL) || (sed==NULL) || (nsed<=0)) {
		nr::nrerror("ATMOS: sedcrv(): null input data") ;
		return ;
	}
	
	/* Find the starting point of the response curve at SED wavelength 
		grid: */
	int k = -1 ;	/* Set "no match" */
	for (int j=0; j<=nsed-nresponse; j++) 
		if (lambda[0]==lambda_sed[j]) { 
			k = j ; break; 
		}

	/* Multiply the arrays, given their different origin */
	if (k>=0)
		for (int j=0; j<atm::nresponse; j++)
			/* Check the match of grids: */
			if (fabs(lambda_sed[j+k] - lambda[j])>wf::EPSLEFF) {
				k = -1;
				break ;
			} else 
				/* k is not negative: no data loss by overwrite below */
				sed[j] = sed[j+k] * response[j] ;

	if (k<0) 
		nr::nrerror("ATMOS: sedcrv(): no match of response and spectrum grids",
			nr::ERCRV) ;

/*	for (j=0; j<nresponse; j++)
		printf("%5.3f %6.4f\n",lambda[j],sed[j]); */
}

/*---------------------------- loadsed --------------------------------*/

int atm::loadsed(const char * sedfile) {
const int MKM_RANGE	= 0	;

	if (nr::erget()) return nr::erget() ;

	/* Check if the response curve is already loaded */
	if ((atm::nresponse<=0) || !atm::lambda) {
		nr::nrerror("ATMOS: loadsed(): null reference inso (is response file "
		"loaded?)",nr::ERNUL) ;
		return nr::ERNUL;
	}
	
	/* Check that spectrum that we read from file is Ok: */
	double * lambda = 0, * sed = 0 ;
	int nsed = 0 ;
	read2col(sedfile, &lambda, &sed, &nsed) ;

	/* Convert wavelengths into [mkm] */
	conwgrid(lambda, nsed, MKM_RANGE) ;

	/* Check that response curve wavelength grid complies the spectrum and
	multiply them: */
	sedcrv(lambda, sed, nsed) ;
	
	/* We do not need SED's wavelength grid since it matches the response grid*/	
	nr::free_dvector(lambda,0,nsed-1) ;
	if (nr::erget()) {
		/* Release test-spectrum and lambda-grid: */
		nr::free_dvector(sed,0,nsed-1) ;
	} else {
		/* Update the global storage replacing the current SED with a new one */
		if (atm::sed)
			nr::free_dvector(atm::sed, 0, atm::nsed) ;
		atm::sed = sed ;
		atm::nsed = nsed ;
	}
	return nr::erget() ;
}

/*------------------------------ checkwf -------------------------------*/

int atm::checkwf(const char * wfile) {

	if (nr::erget()) return nr::erget() ;

	if (!wfile) {
		nr::nrerror("ATMOS: checkwf() on null file", nr::ERNUL) ;
		return nr::ERNUL;
	}
	
	/* Compute and store the current spectrum characteristics in wf: */
	atwf.calcleff(lambda, sed, nresponse, wf::LEFFFIX) ;

	/* Check if the weight file wfile fits the current state of wf: */
	return atwf.checkfile(wfile) ;
}

/*------------------------------ calcwf --------------------------------*/

void atm::calcwf(const char * wfile, void(* progress)(int)) {

	if (nr::erget()) return ;
	
	/* Check is the problem the Z-grid: */
	int err = checkwf(wfile) ;

	if (err<nr::ERFATAL)
		nr::erreset() ;/* reset non-fatal error */
	
	if (err!=nr::ERZGR) {
		/* Problem was not in the altitude grid: compute with
		default optimal grid : */
		atwf.clear() ;
		atwf.setzgrid(wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN) ;
	} /* else - Leave the user's grid intact */
	/* Since sed is shifted to the origin of response curve, Nsed =
			nresponse : */
	if (!nr::erget())
		atwf.calc(atm::lambda, atm::sed, atm::nresponse, progress) ;

	if (!nr::erget())
		/* Write the new weight file: */
		atwf.write(wfile) ;
}

/*------------------------------ update --------------------------------*/

void atm::update(
	const char * wfile,
	const char * checkfile,
	double zshift1,
	double zshift2) {

const double KM2M = 1e3 ;

	/* User must put the SED and weight file names to read. Although they may
	be the same as in previous run and cause no action at all */
	
	if (nr::erget()) return ;

	if (!wfile) {
		nr::nrerror("ATMOS: update() on null spectrum file name", nr::ERNUL) ;
		return ;
	}
	
	/* Check the weight to coincide with the currently loaded by its file name */
	if (curwffile != string(wfile)) {
		/* weight is either not loaded or different */
		curwffile = wfile ;

		/* Check and read (if ok) the file with weight: */
		int err = atm::checkwf(wfile) ;
		if (err) {
			string msg = "ATMOS: update(): checkfile(" ;
			msg += wfile;
			msg +=") returned: " ;
			msg += nr::ercodemessage(err) ;
			nr::nrerror(msg.c_str(),err) ;
			return ;
		} else
			atwf.read(wfile) ;
	} else {
		/* Weight is the same. Check for the shifts and exit if they are so: */
		if ((zshift[0]==zshift1) & (zshift[1]==zshift2)) 
			return ;
	}
	if (nr::erget()) return ;	/* returns if either checkwf() or wf_t::read()
								got bad */
	
	/* Either weight functions or shifts have changed: redo decomposition */

	/* Cycle with 2 steps: 0-th for moments decomposition for the normal mode;
	   1-st for generalized (2 zshifts) mode */

	/* Check-file opeing: */
	FILE * cf = 0 ;
	if (checkfile)
		if (!(cf=fopen(checkfile,"wt"))) {
			nr::nrerror((string("ATMOS: update(): problem to open the "
			"check-file ")+checkfile).c_str(), nr::ERFIO) ;
			return ;
		}

	double ** wmatr=0, ** wcopy=0, ** vmatr=0;
	double  * wvect=0, * hvect=0 ;	/* SVD spaces */
	int ibound = 0 ;	/* index of the boundary layer limit in WF z-grid */
	int nz = 0;
	for (int gen=0; gen<NMODE; gen++) {

		double curzshift = (gen ? zshift2 : zshift1) ;
		zshift[gen] = curzshift ;


		/* COMPUTE THE COMBINED WEIGHT MATRIX FOR Cn2 RESTORATION with dz=1km */

		/* (gen+1)-th part of this matrix is obtained now: */
		wf_t wfs ;
		wfs.setzgrid(floor(z0),floor(zmax),DZCN2,0) ; 

		wfs.interpolate(&atwf, curzshift) ;
		if (nr::erget()) return ;

		/* Fill in the respective part of combined weight matrix */
		if (cf) {
			fprintf(cf,"# Use weights:") ;
			if (gen) fprintf(cf," non-shifted +");
		}
		for (int j=0, k=gen*atm::nweight[0]; k<nweight[gen]; j++) {
			if ((!gen) || (atm::USESHFT[j]!='0')) {
				wfs.copy(0, wfs.getnz()-1, j, wcomb[k]) ;
				if (cf) {
					if (!gen) fprintf(cf," %s",wfs.name(j)) ;
					else fprintf(cf," %ss",wfs.name(j)) ;
				}
				k++ ;
			}
		}
		
		if (cf) fprintf(cf, "\n") ;
		/* Clear shifted weight back: */
		wfs.setaper(0) ;

		/* COMPUTE THE TRANSPOSED WEIGHT MATRIX FOR ALTITUDE POWERS
		DECOMPOSITION */
		wfs.setzgrid(z0,zmax,dz,dzmin) ; 
		wfs.interpolate(&atwf, curzshift) ;
		if (nr::erget()) return ;
		
		if (!gen) {
			/* First loop: allocate memory with just got known dimensions */
			nz = wfs.getnz() ;
			wmatr=nr::dmatrix(1,nz,1,atm::nweight[1]) ;
			wcopy=nr::dmatrix(1,nz,1,atm::nweight[1]) ;
			wvect=nr::dvector(1,atm::nweight[1]) ;
			vmatr=nr::dmatrix(1,atm::nweight[1],1,nweight[1]) ;
			hvect=nr::dvector(0,nz-1) ;

			/* Find the zgrid-index of the boundary layer end: */
			for (int j=0; j<nz; j++) {
				if (wfs.getalt(j)>=atm::HBOUND) {
					ibound = j;
					break ;
				}
			}
		} else {
			/* Restore the wmatr from the copy to update its right-side col-s */
			for (int j=0; j<atm::nweight[0]; j++) 
				for (int i=0; i<nz; i++) 
					wmatr[i+1][j+1] = wcopy[i+1][j+1] ;
		}

		if (ibound<=0) {
			nr::nrerror("ATMOS: update(): failed to find the boundary layer"
			" index",nr::ERZGR); 
			goto ENDLAB ;
		}
		/* Combined transposed matrix of weights for SVD */
		for (int j=0, k=gen*nweight[0]; k<nweight[gen]; j++) {
			if ((!gen) || (atm::USESHFT[j]!='0')) {
				for (int i=0; i<nz; i++) 
					wmatr[i+1][k+1] = wcopy[i+1][k+1] = wfs.getval(i,j) ;
				k++ ;
			}
		}

		/* Checking for permutation invariancy: */
/*		for (int j=0;j<nz;j++) { d = wmatr[j+1][5];	wmatr[j+1][5] = wmatr[j+1][16] ;
			wmatr[j+1][16] = d;	}*/
	
		/* Run SVD of weight matrix: */
		nr::svdcmp(wmatr,nz,nweight[gen],wvect,vmatr) ;
		if (nr::erget()) goto ENDLAB ;

		/* scalar product check: */
/*		for (int i=0;i<nweight[gen]-1;i++) {
			d = 0; for (int j=0;j<nz;j++) d+= wmatr[j+1][i+1]*wmatr[j+1][i+2] ;
			printf("%d-%d %10.4e\n",i,i+1,d); }
*/
		/* Maximal singular value: */
		double d = wvect[1] ;
		for (int i=2;i<=atm::nweight[gen];i++) 
			if (wvect[i]>d) d=wvect[i] ; /* maximal w */
		/* Minimal acceptable singular value: */
		d *= atm::MINRELW ;	

		int nbase = nweight[gen] ;
		/* Rejection of small singulars: */
		if (cf) fprintf(cf, "# N(Base functions): Init.=%d, ",nbase) ;
		for (int i=1;i<=nweight[gen];i++) 
			if (wvect[i]<d) { 
				wvect[i] = 0 ; 
				nbase--; 
			}
		if (cf) fprintf(cf, " Smalls reject=> %d, ",nbase);
		/* Rejection of least singulars: provided they are at the end of 
			wvect[] */
		for (int i=atm::nweight[gen];(i>0) && (nbase>atm::MAXNBASE);i--) 
			if (wvect[i]!=0) {
				wvect[i] = 0 ; 
				nbase--; 
			}
		if (cf) 
			fprintf(cf, " Restricted to => %d\n# Approximated powers: ",nbase);
	
		for (int ipow=int(atm::IFSE); ipow<=int(atm::IISK); ipow++) {
			/* Select the power: */
			double power ;
			switch (poweridx(ipow)) {
				case IFSE:
				case IWSE: power = POWSEE; break ;
				case IISP: power = POWISP; break ;
				case IEFF: power = POWEFF; break ;
				case IISK: power = POWISK; break ;
				default: nr::nrerror("ATMOS: update(): invalid power index",
				nr::ERPAR) ;
				goto ENDLAB ;
			}
			if (cf) fprintf(cf, "\n# Ap%d: %6.3f",(int)ipow,power) ;
			/* Generate the altitude powers: */
			for (int j=0; j<nz; j++) 
				hvect[j] = pow(wfs.getalt(j)*KM2M, power) ;
			if (ipow==int(atm::IFSE)) {
				/* free atmosphere: force the hvect to resemble the smallest
				aperture pair weight behavior: */
				wfs.copy(0,ibound-1,nchan,hvect) ;
				d = wfs.getval(ibound,nchan) ;
				for (int j=0; j<ibound; j++) hvect[j] /= d ;
				if (cf) fprintf(cf," (W(AB)-dumped)");
			}
			/* Check the powers: */
/*if (ipow==IFSE) for (j=0;j<nz;j++) printf("%f %f\n",wf_getalt(&wfs,j),hvect[j]);*/
			/* Back-substitute to find the coefficients: */
			nr::svbksb(wmatr,wvect,vmatr,nz,atm::nweight[gen], hvect-1, 
				atm::coef[gen][ipow]-1);
		
		}
		if (cf) {
			/* Check the coefficients: */
			fprintf(cf,"\n# Coefficients:\n");
			for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
				/* Maximal one: */
				d = 0 ;
				for (int j=0; j<nweight[gen]; j++) {
					double ac = fabs(coef[gen][ipow][j]) ;
					if (d<ac) d=ac ;
				}
				fprintf(cf,"# Ap%d: Max=%9.3e",ipow,d);
				for (int j=0; j<atm::nweight[gen]; j++) 
					fprintf(cf," %7.4f",atm::coef[gen][ipow][j]/d);
				fprintf(cf,"\n");
			}
			/* Check the approximated relations: */	
			fprintf(cf,"# h[km]  ") ;
			for (int i=0; i<=atm::IISK ; i++) 
				fprintf(cf,"#Ap%d/h^ ",i) ;
			fprintf(cf,"\n") ;
			for (int i=1;i<nz;i++) {
				fprintf(cf,"%6.3f",wfs.getalt(i)) ;
			   for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
			   		double power ;
					switch (poweridx(ipow)) {
						case IFSE:
						case IWSE: power = POWSEE; break ;
						case IISP: power = POWISP; break ;
						case IEFF: power = POWEFF; break ;
						case IISK: power = POWISK; break ;
						default: goto ENDLAB ;
					}
					/* power of altitude: */
					curzshift = pow(wfs.getalt(i)*KM2M,power) ;
					d = 0 ;
					for (int j=0; j<atm::nweight[gen]; j++) 
						d+=wcopy[i+1][j+1]*coef[gen][ipow][j];
					fprintf(cf," %6.3f",d/curzshift); 
				}
				fprintf(cf,"\n") ;
			}
		}
		/* Clear the shifted weight */
		wfs.setaper(0) ;
	}

	/* Close check-file */
	if (checkfile)
		fclose(cf) ;

ENDLAB:
	/* Free the orthogonalized matrix, singular value vector and square 
	V-matrix: */
	nr::free_dmatrix(wmatr,1,nz,1,atm::nweight[1]) ;
	nr::free_dmatrix(wcopy,1,nz,1,atm::nweight[1]) ;
	nr::free_dvector(wvect,1,atm::nweight[1]) ;
	nr::free_dmatrix(vmatr,1,atm::nweight[1],1,atm::nweight[1]) ;
	nr::free_dvector(hvect,0,nz-1) ;

	/* reset counter of made measurements of integrals: */
	curmeas = 0 ;

}

/*------------------------------ alloc --------------------------------*/

void atm::alloc(int nmeas) {
	/* Allocate the storages for integrals: */
	if (nmeas>maxnmeas) {
		if (atm::maxnmeas>0) {
			nr::free_dmatrix(cn2int,0,atm::maxnmeas-1,0,LAST-1) ;
			nr::free_dvector(aint,0,LAST-1) ;
			nr::free_dvector(e2int,0,LAST-1) ;
		}
		cn2int=nr::dmatrix(0,nmeas-1,0,LAST-1) ;
		aint=nr::dvector(0,LAST-1) ;
		e2int=nr::dvector(0,LAST-1) ;
		atm::maxnmeas = nmeas ;
	}
	curmeas = 0 ;
}

/*------------------------------ done --------------------------------*/

inline void debug(const char* s) {
//	printf("ATM:---------------%s----------------\n",s);
}

void atm::done(void) {
debug("atm::done()");
	if (atm::sed) {
		debug("delete sed[]");
		nr::free_dvector(atm::sed,0,atm::nsed) ;
		atm::sed = 0 ;
		atm::nsed = 0 ;
	}
	
	if (atm::nresponse>0) {
		debug("delete response[]");
		nr::free_dvector(response,0,nresponse-1);
		nr::free_dvector(lambda,0,nresponse-1);
		nresponse = 0 ;
		response = lambda = 0 ;
	}
	if ((nweight[1]>0) && (wcomb)) {
		debug("delete wcomb[][]");
		nr::free_dmatrix(wcomb,0,nweight[1]-1,0,0) ;
		wcomb = 0 ;
	}
	if (ncn2>0) {
		debug("delete Cn2[]");
		nr::free_dvector(cn2,0,ncn2);
		nr::free_dvector(zcn2,0,ncn2);
		ncn2 = 0 ;
		cn2 = zcn2 = 0 ;
	}
	atwf.setaper(0) ;
	atm::nchan = 0 ;
	
	if (atm::coef){ 
		debug("delete coef[][]");
		nr::free_dmatrix(coef[0],0,NPOWER-1,0,NPOWER); /* last par. is foo*/
		nr::free_dmatrix(coef[1],0,NPOWER-1,0,NPOWER) ;
		coef[0] = coef[1] = 0 ;
		nweight[0] = nweight[1] = 0 ;
	}

	/* Deallocate the storages for integrals: */
	if (cn2int!=NULL) {
		debug("delete cn2int[][]");
		nr::free_dmatrix(cn2int,0,maxnmeas-1,0,LAST-1) ;
		nr::free_dvector(aint,0,LAST-1) ;
		nr::free_dvector(e2int,0,LAST-1) ;
		cn2int = NULL ; aint = e2int = NULL ;
		maxnmeas = curmeas = -1 ;
	}
	debug("done Ok");
}

/*------------------------------ calcint --------------------------------*/

void atm::calcint(double * scind, double * scinds) {

	if (!coef[0] || !scind) {
		nr::nrerror("ATMOS: calcint() on empty index or coef. array", nr::ERNUL) ;
		return ;
	}
	
	if (curmeas>=maxnmeas) {
		nr::nrerror("ATMOS: calcint() overflows the storage", nr::EROFL) ;
		return ;
	}

	/* We detect the generalized mode by the appearance of non-NULL scinds */
	bool isgen = (scinds!=0) ;

	/* Compute the Cn2-integrals from indices: */
	for (int i=0; i<NPOWER; i++) {
		double cn2mom = 0 ;
		for (int j=0; j<nweight[0]; j++) 
		 	cn2mom += scind[j] * coef[isgen][i][j] ;
		if (isgen) {
			for (int j=0, k=atm::nweight[0]; k<nweight[1]; j++) 
				if (atm::USESHFT[j]!='0') {
				 	cn2mom += scinds[j] * coef[isgen][i][k] ;
					k++ ;
				}
		}
		cn2int[curmeas][i] = cn2mom ;
	}
	curmeas++ ;
}


/*---------------------------- meritfunc --------------------------------*/
/* Reopen atm once again: */
namespace atm {

int curnweight = 0 ;
/**
	@brief	Merit function for the profile-restoration at fixed altitudes
	@param	y	vector of sqrt(Cn2) of layers
	@return	merit Chi-square
	
*/
double meritfunc(double *) ;

const int NFIXED = 6 ;	/* Number of fitted layers with fixed altitudes (-1 in
normal mode) */
const int NFLOAT = 4 ;	/* ditto with adjusted altitudes */

}

double atm::meritfunc(double * y) {
	double chi2 = 0 ;	/* Khi-square */
	for (int j=0; j<curnweight; j++) {
		double sc2 = 0 ;	/* scint. index */
		for (int i=0; i<ncn2; i++) 
			sc2 += y[i+1] * y[i+1] * wcomb[j][int(rint(zcn2[i]))] ;
		double d = sc2 - sicomb[j] ;
		chi2 += d*d / e2sicomb[j] ;
	}
	return chi2 ;
}

/*---------------------------- calcn2 ----------------------------------*/

double atm::calcn2(double * scind, double * scinds,
		double * e2scind, double * e2scinds, bool isfixed) {
const double ZFIXFACT = 2.0	;/* Factor of altitude modification in 
fixed-altitude method, should be about 1/(alt.resolution) */
const int ZGND = 0 ;	/* integer-type altitude of ground (layer) */
const int ZBOUND = 1 ;	/* integer-type altitude of a boundary layer */
const double AVGCN2	= 6.95e-13 ;	/* Cn2-integral giving 1" seeing */
const double FTOL = 1e-3 ;		/* Tolerance in powell() */

const int MAXH1 = 5 ;	/* Maximal altitude of the second floating layer */
const int MAXH2 = 12 ;	/* Maximal altitude of the third floating layer */
const int MINDH = 1 ;	/* Minimal altitude difference between layers */

	if (nr::erget()) return bad;

	/* Combine the indices: */
	if (!scinds) 
		atm::curnweight = nweight[0] ;
	else
		atm::curnweight = nweight[1] ;

	atm::sicomb =   nr::dvector(0,curnweight-1) ;
	atm::e2sicomb = nr::dvector(0,curnweight-1) ;
	for (int i=0; i<atm::nweight[0]; i++) {
		sicomb[i] = scind[i] ;
		e2sicomb[i] = e2scind[i] ;
	}
	for (int i=0, k=nweight[0]; k<curnweight; i++) {
		/* works in generalized mode only */
		if (atm::USESHFT[i]!='0') {
			sicomb[k] = scinds[i] ;
			e2sicomb[k] = e2scinds[i] ;
			k++ ;
		}
	}

	/* Allocate maximal of possibly needed memory if not allocated */
	if (!cn2) {
		int i = (NFIXED>NFLOAT ? NFIXED : NFLOAT) ;
		/* Allocate longest needed vectors */
		zcn2=nr::dvector(0,i-1) ;
		cn2=nr::dvector(0,i) ;	/* +1 element for Chi-2 */
	}

	double minchi2 = 1/FTOL/FTOL ;	/* arbitrary large number */
	if (isfixed) {
		/* Fixed layers method */
		/* Generate the grids: altitudes and 1st approximation of the profile */
		atm::ncn2 = NFIXED - (scinds==0);
		atm::zcn2[0] = ZGND + (double)(scinds==0);	/* ground layer or bnd.lay.*/
		atm::zcn2[1] = ZBOUND + (double)(scinds==0);
		for (int i=2; i<ncn2; i++) {
			atm::zcn2[i] = atm::zcn2[i-1]*ZFIXFACT ;
			atm::cn2[i] = sqrt(AVGCN2 / atm::ncn2) ;
		}
		/* Diagonal "identity" matrix: */
		double ** xi = nr::dmatrix(1,atm::ncn2,1,atm::ncn2) ;
		for (int i=1; i<=ncn2; i++) 
			for (int j=1; j<=ncn2; j++) xi[i][j] = (double)(i==j) ;
			
		/* Solve the task: find the minimum chi-square */
		int iter = 0 ;
		nr::powell(cn2-1,xi,ncn2,FTOL,&iter,&minchi2,atm::meritfunc);
		/* Back-square the profile: */
		for (int i=0; i<ncn2; i++) cn2[i] *= cn2[i] ;
		cn2[ncn2] = minchi2 ;
		nr::free_dmatrix(xi,1,ncn2,1,ncn2) ;
	} else {
		/* "Floating" layers method */
		atm::ncn2 = NFLOAT ;
		double **w3=nr::dmatrix(1,curnweight,1,ncn2) ;
		double **w3inv=nr::dmatrix(1,ncn2,1,curnweight) ;
		double *wvect=nr::dvector(1,ncn2) ;
		double **vmatr=nr::dmatrix(1,ncn2,1,ncn2) ;
		double * tr_cn2=nr::dvector(0,ncn2-1) ;// trial Cn2 vector
		int ih[NFLOAT] = {(scinds==NULL)} ; /* Fixed ground layer */
		for (ih[1]=ih[0]+MINDH; ih[1]<=MAXH1; ih[1]++)
			for (ih[2]=ih[1]+MINDH; ih[2]<=MAXH2; ih[2]++)
				for (ih[3]=ih[2]+MINDH; ih[3]<=(int)atm::zmax; ih[3]++) {
					/* Compose the selected weights matrix */
					for (int j=0;j<curnweight;j++)
						for (int i=0;i<ncn2; i++)
							w3[j+1][i+1] = wcomb[j][ih[i]] ;

					/* Invert w3-matrix */
					nr::svdcmp(w3,curnweight,ncn2,wvect,vmatr) ;
					/* Cancel small singulars */
					double d = 0 ;
					for (int i=1; i<=ncn2; i++)	if (d<wvect[i]) d=wvect[i];
					d *= MINRELW ;	/* Minimal acceptable singular value */
					for (int i=1; i<=ncn2; i++)
						if (wvect[i]>d) 
							wvect[i] = 1. / wvect[i] ;
						else 
							wvect[i] = 0 ;
					/* multiply three matrices: A^{-1}=(V*diag(1/w))*U^T */
					for (int i=1; i<=ncn2; i++) 
						for (int j=1; j<=curnweight; j++) {
							w3inv[i][j] = 0 ;
							for (int k=1; k<=ncn2; k++)
								w3inv[i][j] += vmatr[i][k]*wvect[k]*w3[j][k] ;
						}

/* Check the inversion: reproduce and display w3, display w3inv and their
product: */
/*
				if (ih[1]==5 && ih[2]==12 && ih[3]==20) {
					for (int j=0;j<curnweight;j++) {
						for (int i=0;i<ncn2; i++) {
							w3[j+1][i+1] = wcomb[j][ih[i]] ;
							printf("%6.3f ",1e-11*w3[j+1][i+1]);
						}
						printf("\n");
					}
					printf("\n");
					for (int i=0;i<ncn2; i++) {
						for (int j=0;j<curnweight;j++) {
							printf("%6.3f ",1e+11*w3inv[i+1][j+1]);
						}
						printf("\n");
					}
					printf("Ok\n");
					for (int i=1;i<=ncn2; i++) {
						for (int j=1; j<=ncn2; j++) {
							double d = 0;
							for (k=1; k<=curnweight; k++)
								d+=w3inv[i][k] * w3[k][j];
							printf(" %5.2f",d);
						}
						printf("\n");
					}
				}
*/
					/* Compute the strengths: */
					for (int i=0; i<ncn2; i++) {
						tr_cn2[i] = 0 ;
						for (int j=0; j<curnweight; j++)
							tr_cn2[i] += w3inv[i+1][j+1] * sicomb[j] ;
						if (tr_cn2[i]<0) tr_cn2[i] = 0 ;	  /* >=0 constraint */
					}

					/* Compute indices and chi-square: */
					double chi2 = 0 ;
					for (int j=0; j<curnweight; j++) {
						double sc2 = 0 ;
						for (int i=0; i<ncn2; i++)
							sc2 += tr_cn2[i] * wcomb[j][ih[i]] ;
						double d = sc2 - sicomb[j];
						chi2 += d*d / e2sicomb[j] ;
					}
					/* Save the altitudes and strengths if this is the best
						case up to now */
					if (chi2<minchi2) {
						minchi2 = chi2 ;
						for (int i=0; i<ncn2; i++) {
							zcn2[i] = (double)(ih[i]) ;
							cn2[i] = tr_cn2[i] ;
						}
					}

				}
		nr::free_dmatrix(w3,1,curnweight,1,ncn2) ;
		nr::free_dmatrix(w3inv,1,ncn2,1,curnweight) ;
		nr::free_dvector(wvect,1,ncn2) ;
		nr::free_dmatrix(vmatr,1,ncn2,1,ncn2) ;
		nr::free_dvector(tr_cn2,0,ncn2-1) ;
		cn2[ncn2] = minchi2 ;	/* save chi-square in last element */
	}
	nr::free_dvector(sicomb,0,curnweight-1) ;
	nr::free_dvector(e2sicomb,0,curnweight-1) ;
	return minchi2 ;
}

/*------------------------------ avgint ----------------------------------*/

/** Compute the formal relative error of the value of the power function of the
value \em mean given its squared error \em e2 */
inline double relerr(double mean, double e2, double power) {
	if ((e2<=0) || (mean<=0)) 
		return 0 ;
	else
		return fabs(power * sqrt(e2) / mean) ;
}

void atm::avgint(double desi, double e2desi, double zdist) {
const double JPOW_SEE = 0.6 ;	/* Power of Cn-integral for seeing */
const double LPOW_SEE = -0.2 ;	/* Power of lambda for seeing */
const double JPOW_ISP = -0.6 ;	/* Power of Cn-5/3th moment for isoplan. angle*/
const double LPOW_ISP = 1.2 ;	/* Power of lambda for isoplan. angle */
const double SPOW_TC  = -0.6 ; /* Power of DESI for atm. time constant */

	if (curmeas<1) {
		nr::nrerror("ATMOS: avgint() with no measurements done",nr::ERPAR) ;
	}

	/* average integrals: */
	sc::avgmatrix(cn2int,LAST,curmeas,0,0,aint,e2int,sc::MAXLAG) ;

	if (nr::erget()) return ;	/* for both two checks and avgmatrix() error */

	/* Convert integrals into atmospheric parameters */
	
	/* secans of the zenith distance [deg] */
	double secz = 1./cos(zdist*M_PI/180.) ;

	/* Seeing : free and whole atmosphere */	
	fseeing = seeing = efseeing = eseeing = bad ;
	int isbadint = false ;
	if (aint[IFSE]>0) {
		double zpow = secz ;	// secz^(POWSEE+1)
		fseeing = KS * pow(aint[IFSE] / zpow, JPOW_SEE) * 
			pow(atwf.getleff(),LPOW_SEE) ;
		efseeing = fseeing * 
			relerr(aint[IFSE],e2int[IFSE],JPOW_SEE);
		seeing = KS * pow(aint[IWSE] / zpow, JPOW_SEE) * 
			pow(atwf.getleff(),LPOW_SEE) ;
		eseeing = seeing * 
			relerr(aint[IWSE],e2int[IWSE],JPOW_SEE);
	} else {
		isbadint = true ;
	}
	
	/* Isoplanatic angle: */
	isoplan = eisoplan = bad ;
	if (aint[IISP]>0) {
		double zpow = pow(secz,POWISP+1) ;
		isoplan = KP * pow(aint[IISP] / zpow, JPOW_ISP)
			* pow(atwf.getleff(), LPOW_ISP) ;
		eisoplan = isoplan *
			relerr(aint[IISP],e2int[IISP],JPOW_ISP); 
	} else {
		isbadint = true ;
	}

	/* Effective altitude of turbulence in [meters]: (free atm.) */
	fheff = heff = efheff = eheff = bad ;
	if (aint[IEFF]>0 && aint[IISP]>0) {
		double power = 1.0/(POWISP-POWEFF) ;
		double zpow = pow(secz, POWISP-POWEFF) ;
		fheff = pow(aint[IISP]/aint[IEFF]/zpow,power) ;
		double d = relerr(aint[IISP],e2int[IISP],power) ;
		efheff = d*d;
		d = relerr(aint[IEFF],e2int[IEFF],power) ;
		efheff += d*d;
		efheff = sqrt(efheff) * fheff ;
	} else {
		isbadint = true ;
	}
	
	/* Whole atm. effective altitude: */
	if (aint[IEFF]>0 && aint[IWSE]>0) {
		heff = aint[IEFF]/aint[IWSE] / secz ;
		double d = relerr(aint[IEFF],e2int[IEFF],POWEFF) ;
		eheff = d*d;
		d = relerr(aint[IWSE],e2int[IWSE],POWEFF) ;
		eheff += d*d;
		eheff = sqrt(eheff) * heff ;
	} else {
		isbadint = true ;
	}

	timeconst = etimeconst = bad ;
	if (desi>0) {
		timeconst = KT * pow(desi,SPOW_TC) ;
		double d = (e2desi>0) ? sqrt(e2desi) : 0 ;
		etimeconst = timeconst * relerr(desi,d,SPOW_TC) ;
	} else {
		isbadint = true ;
	}
	if (isbadint) {
		nr::nrerror("ATMOS: at least one Cn2-integral is non-positive. "
		"Skip writing the results",nr::ERNOD) ;
	}
}

/*----------------------------- writeint -------------------------------*/

void atm::write(FILE * f, const char * stamp, atmwhat what, 
	bool header = false) {

	if (!f) nr::nrerror("ATMOS: write() with null FILE", nr::ERPAR) ;
	if (nr::erget()) return ;
	
	/* write the stamp: */
	if (stamp)
		if (!fprintf(f, "%s",stamp)) {
			nr::nrerror("ATMOS: write(): cannot write file", nr::ERFIO) ;
			return ;
		}
		
	switch (what) {
		case ATMPAR:
		if (header) {
			for (int i=0; i<int(atm::LAST); i++) {
				char * hl ;
				switch (atm::ordint(i)) {
					/* headings with TWO %s for TWO delimiters: */
					case FSEE:	hl="%s fSee %se_fSee"; break;
					case WSEE:	hl="%s  See %s e_See"; break;
					case ISP:	hl="%s Isop %se_Isop"; break;
					case M2:	hl="%s  Cn2h2 %s e_Cn2h2"; break;
					case FHEFF:	hl="%sfHeff%se_fHe"; break;
					case HEFF:	hl="%s Heff%s e_He"; break;
					case TC:	hl="%s TAU %se_TAU"; break;
					default:	hl="";
				}
				fprintf(f,hl,atm::INTDLM,atm::INTDLM) ;
			}			
		} else {
			/* write the average integrals and their errors: */
			for (int i=0; i<int(atm::LAST); i++) {
				double d = getval(ordint(i)); 
				double e = geterr(ordint(i)) ;
				fprintf(f,INTFMT[i],atm::INTDLM,d) ;
				fprintf(f,INTFMT[i],atm::INTDLM,e) ;
			}
		}
		break;
		case CN2PROF:
			if (header) {
				fprintf(f, "%sNz%s Chi2 ",INTDLM, INTDLM) ;
				for (int i=0; i<NFIXED; i++)
					fprintf(f,"%sz%d%s  Cn2_%d ",INTDLM,i+1,INTDLM,i+1) ;
			} else {
				/* save the dimention, altitudes and strengths of the restored 
					profile: */
				fprintf(f,"%s%2d%s%6.2f",INTDLM, getncn2(),INTDLM, 
					getcn2(getncn2())) ;
				for (int i=0; i<getncn2(); i++) {
					fprintf(f,ALTFMT,INTDLM,getzcn2(i)) ;
					fprintf(f,CN2FMT,INTDLM,getcn2(i)) ;
				}
			}
	}
	/* Set the carriage-return character: */
	if (!fprintf(f,"\n"))
		nr::nrerror("ATMOS: write(): cannot write file 2", nr::ERFIO) ;
}
