/*--------------------------------------------------------------------------*/
/** @file atmos.hxx
                  MASS project: TURBINA module file                       
                       header file for atmos.cpp                             
                                                                          
 The module ATMOS specifies the top-level calculations which are aimed to
 compute the atmospheric parameters and the turbulence profile. The weight
 functions, described themselves in the module WEIF, are selected and 
 initialized here. With the help of these functions, the various integrals over
 the altitude which approximate the certain moments of the turbulence
 (calibrated into seeing, isoplanatic or isokinetic angles etc) are calculated
 and the low-resolution turbulence profile is computed.

 \b Usage:

 Prior to the scientific operations using the observation data, the module
 should be initialized with atm::init(). Since the module acts as an interface
 to the module WEIF, some parameters for weight initialization (which are not
 changed frequently) are transmitted via atm::init(), some others (such as
 number of measurements or spectral energy distribution - via atm::update().
 Note, that  the initialization of the module SCIND has to be done separately.
 Each time some basic information is changed (apertures geometry, transmission
 function, altitude grid), the module has to be shut down with atm::done() and
 reinitialized with atm::init() again.

 Each time the new object is selected, the weight functions must be updated
 using the object's own spectral energy distribution (SED) file via the call to
 atm::loadsed(), atm::checkwf() and atm::update() functions. 

 Before starting the new measurement the function atm::alloc() must be called \b
 before the first calculation of atmospheric integrals. atm::alloc() requires
 the guessed number of index measurements per accumulation time. It must not be
 less than actual to avoid the error on the storages overfull.

 Then, after some (instantaneous) scintillation indices are obtained, the
 computation of integrals of Cn2*h^a over atmosphere (power a=0,1,5/3,2) has to
 be done with atm::calcint(). The results are stored internally.

 After the completion of the accumulation time, the integrals of Cn2 by powers 
 of altitude are averaged and converted into seeing, isoplanatic angle and
 effective altitude of turbulence by atm::avgint(). Also, the procedure 
 atm::calcn2() may be called which uses the average indices and their errors to 
 restore the low-resolution Cn2-profile with one of two methods. 

 Once the profile and atmospheric characteristics are computed, they may be
 saved on disk by atm::write(). The converted average  integral values and
 their errors may be accessed by atm::getval() and atm::geterr() after
 atm::avgint() call.

 Finally, to end with the measurements, call a memory destructor atm::done().

 @version 1.5: Translation into C++                                         */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _ATMOS_H
#define _ATMOS_H

#include <stdio.h>

/**
	@brief
	Namespace at contains the set of functions of highest logical level for
	atmospheric calculations
*/
namespace atm {

/** Order of atmosphere integrals in their storage matrix and output file line.
In a file, each value is followed by its error. LAST must stay always 
at the end since it gives a number of integrals in enumeration */
enum ordint {
	FSEE,	/**< free seeing */
	WSEE,	/**< whole atm. seeing */
	ISP,	/**< isoplanatic angle */
	M2,		/**< second moment of turbulence (integral(Cn2(h)*h^2dh)) */
	FHEFF,	/**< effective altitude of turbulence in free atm. */
	HEFF,	/**< effective altitude of turbulence in whole atm. */
	TC,  	/**< atmospheric time constant */
	LAST
} ;

/** Writing selector in write(): */
enum atmwhat {
	ATMPAR,	/**< write atmospheric parameters */
	CN2PROF	/**< write the restored Cn2 profile with altitudes and Chi-square */
} ;

/** Last (defaulted to "data"==false) parameter in write() */
const bool HEAD = true ;	

/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
	@brief	Initialize the module local structures for calculations 
	@param	nchan	Number of the MASS detector channels to use
	@param	diam	see wf_t::setaper()
	@param	eps_inn	see wf_t::setaper()
	@param	eps_out	see wf_t::setaper()
	@param	z0  	see wf_t::setzgrid()
	@param	zmax	see wf_t::setzgrid()
	@param	dz  	see wf_t::setzgrid()
	@param	dzmin	see wf_t::setzgrid()
	@param	responsefile The system transmission function file 
		"lambda[mkm or nm or \AA] P(lambda)"

	This function should be called, in general, once per working session, or
	if some basic system parameters (see the function parameters list) are 
	changed. In this case, done() must be invoked in advance.
	
	First, the global structures are initialized, including the normal and
	shifted weight functions sets. Then the weight-related instrumental
	parameters are initialized. Finally, the system transmission function is
	read from the file. The wavelengths, if different from [mkm], are converted
	into micrometers automatically.
	
	@see done()
*/
void init(
	int nchan, double diam, double *eps_inn, double *eps_out,
	double z0, double zmax, double dz, double dzmin,
	const char * responsefile) ;

/**
	@brief	Load the spectrum from file and check its validity
	@param	sedfile	name of the two-column ASCII file with spectrum 
		"lambda[mkm or nm or \AA] E(lambda)"
	@return	0 if Ok, <0 on error 
		
	The function calls the local file-reader utility (the same as used for the
	MASS  response curve reading). If the file is successfully read (seems that
	spectrum does not contain errors as it is), the wavelengths, if different
	from [mkm], are converted into micrometers. Then an attempt is performed to
	multiply this spectrum by response curve read in init(). Here the
	negative error code may be returned if no match is found between them. If
	the spectrum matches the response curve, it is stored in a global module
	storage to be used in weight functions handling.

	Possible errors (see nr::erget()) include: 
	- sedfile is not readable 
	- init() is not called in advance 
	- spectrum is zero or spectrum does	not match the MASS response curve with 
		any offset. In this case nr::ermessage() may specify the row number of 
		the spectrum file \em sedfile from where the problem originates.

	@see update()
*/
int loadsed(const char * sedfile) ;

/**
	@brief	Check the weight file to fit the currently set system parameters 
	and loaded spectrum
	@param	wfile	weight file name
	@return	0 if file fits, <0 on mismatch or error
	
	The function wf_t::calcleff() is used with the current spectrum and then 
	wf_t::checkfile() is used to check the file. The result of this function is
	returned.

*/
int checkwf(const char * wfile) ;

/**
	@brief	Calculate the new set of weight functions and write it in the file
	@param	wfile	file name to write the weight functions set in
	@param	progress	progress indicator function (may be null)
	
	This function first checks the \em wfile with checkwf(). If result is
	zero (i.e. "ok"), nothing follows. If not, the weights are calculated with
	currently loaded spectrum and system parameters. 

	The grid parameters depend on the result of checkfile(). If the checking
	resulted in error sc::ERZGR (user-supplied grid in init() does not fit the
	file grid), the weights are calculated with user's grid, otherwise - with
	the "optimal" grid. Optimal grid parameters are specified in WEIF module -
	wf::Z0, wf::ZMAX, wf::DZ and wf::DZMIN.

*/
void calcwf(const char * wfile, void(* progress)(int)) ;

/**
	@brief	Update the data structures for atmospheric calculations with a
		new weight function file
	@param	wfile	The file name of the weight matrix (see wf_t::write()) 
	@param	checkfile	file to print the moment approximation information for
		checking (may be null)
	@param	zshift1	Altitude shift (in [km], normally \b non-positive) for the
	normal mode or for \em one part of the generalized mode measurements
	@param	zshift2	Altitude shift (in [km], normally \b negative) for the 
		\em rest part of the generalized mode measurements

	This function should be called each time the star (i.e. its SED) or the 
	measurement mode (generalized or normal; number of base times per 
	accumulation time etc), is changed. 
 
	First, if the weight function matrix is new (its lastly loaded filename is
	stored internally), it is read from the file and all the calculations of
	the Cn2 moments decomposition are made. Also, the calculations are done if
	the z-shifts \em zshift1 and \em zshift2 have changed. Calculations involve
	the shift of weights by \em zshift1 and \em zshift2 and the approximation 
	of the power-law dependences of altitude by shifted weight functions. The
	SVD method is used for this. Approximation enable us to represent the
	Cn2-moments by a set of scintillation indices. If \em zshift1 or \em
	zshift2 are non-zero, then both the coefficients for generalized and
	non-shifted modes are obtained. The coefficients are stored in global
	arrays to be used in calcint().

	The quality of the approximation of the altitude moments with weight
	functions may be checked if one supplies the non-null file name \em
	checkfile. The utility writes then the decomposition coefficients and the
	approximated relations in this file. The file \em checkfile is rewritten
	each time, not appended, and consists of two similar parts - the first for
	approximations with \em zshift1 -shifted weights (normally - non-shifted)
	and the second for approximation with both \em zshift1 and \em zshift2
	-shifted weights.
		
*/
void update(const char * wfile, const char * checkfile, 
	double zshift1, double zshift2) ;
				
/**
	@brief	Allocate the storages for atmospheric integrals calculated by
		calcint()
	@param	nmeas	maximal possible number of calcint() starts (base times
		per accumulation time)
*/
void alloc(int nmeas) ;		

/**
	@brief	Deallocate the memory reserved for module structures
	@return	void
	
	This function releases the memory which is allocated for weights, SED,
	coefficients of Cn2-moments decomposition etc by init().
	
	\note
	This function must be kept in strict sync with init(), update() and alloc().
*/
void done() ;

/****************************************************************************/
/*                    Global atmospheric parameters access                  */
/****************************************************************************/

/** 
	@brief Return calibrated atmospheric parameter
	@param	what	one of computed parameters - seeing, Heff, isoplanatic angle
	etc. presented in ordint enumeration
	@return	value or 0 on error
	
	@see avgint()
*/
double getval(ordint what) ;

/** 
	@brief Return error of the calibrated atmospheric parameter
	@param	what	one of computed parameters - seeing, Heff, isoplanatic angle
	etc. presented in ordint enumeration
	@return	error value or 0 on error
	
	@see avgint()
*/
double geterr(ordint what) ;


/** 
	@brief Return number of the Turbulence profile altitudes 
	@return Length of the private computed Cn2-array 

	Note that this number is not necessary the length of allocated array
	\em cn2[].
	
	@see calcn2()
*/
int getncn2() ;

/** 
	@brief Get the altitude of the restored turbulence profile layer
	@param	i	layer index [0..getncn2()-1]
	@return altitude [km]
	
	@see calcn2()
*/
double getzcn2(int i) ;

/** 
	@brief Get the strength of the restored turbulence profile layer or
	chi-square
	@param	i	layer index [0..getncn2()-1] or getncn2() for Chi-square 
	@return strength [m^-2/3]
	
	Last element of required array contains the Chi-square measure of the fitted
	profile accessible as getcn2(getncn2()).

	@see calcn2()
*/
double getcn2(int i) ;

/****************************************************************************/
/*                     Calculation of atmosphere parameters                 */
/****************************************************************************/

/**
	@brief Compute the atmosphere integral parameters and Cn2-profile
	@param	scind	array of non-shifted indices
	@param	scinds	array of shifted indices (may be NULL)
	
	Using the Cn2-moments decomposition coefficients and  scintillation
	indices, the calculation of integral atmospheric parameters is performed.
	Number of indices in either \em scind or \em scinds must be equal to number
	of respective computed weights in \em wf which are made in init()
	call. No check of the validity of supplied \em scind(s) arrays is done.
	
	The coefficients of multiplication of the indices in normal and shifted
	mode are already precomputed in init(); the results become accessible after
	the function call by getval() and geterr(). The atmospheric time constant
	is computed given the \em desi parameter.  If \em scind(s) or \em desi
	parameters are empty (NULL or 0, respectively), the calculations for
	supplied not-NULL parameters are only made.

	The computed quantities are saved in the current ine of the global 
	storage for further averaging	by avgint(). The number of calls to
	calcint() should not exceed the \em nmeas parameter of alloc().
	
	\b Performance
	
	One calculation of integrals takes \b 5 ms at PC P-III 667 Mhz.

*/
void calcint(double * scind, double * scinds) ;

/**
	@brief Restore the low-resolution Cn2 profile
	@param	scind	array of average non-shifted indices
	@param	scinds	array of average shifted indices (may be NULL)
	@param	e2scind	array of \em squared errors of \em scind
	@param	e2scinds	array of \em squared errors of \em scinds
	@param	isfixed	Use fixed-altitude layers restoration rather than strongest
		layers search
	@return Minimal chi-square reached, or (double)(error_code,<0) on error
	
	The profile restoration results in altitudes of the getncn2() layers 
	altitudes getzcn2() with strengths getcn2(). Two methods are
	implemented - the search for four strongest layers (if not \em isfixed) and
	the search for the intensity of six layers placed proportionally in the
	altitude range with relative resolution 0.5.

	\b Performance
	
	One restoration takes typically \b 45 \em ms at PC P-III 667 Mhz for the
	method with 6 \em fixed layers, and \b 140 \em ms for placing 4 \em 
	floating altitude layers.

*/
double calcn2(double * scind, double * scinds,
	double * e2scind, double * e2scinds, bool isfixed) ;
	
/**
	@brief	Average Cn2 integrals and compute from them the atmospheric 
		parameters
	@param	desi	average DESI for shifted channel A
	@param	e2desi	squared error of \em desi
	@param	zdist	zenith distance in [degree]
	
	Given the instantaneous integrals in their storage, their average values and
	standard deviations are computed and put in respetive global vectors.
	The function sc::avgmatrix() is used for averaging. Errors thus take into
	account the correlation of values.

	Then the average integrals are converted into atmospheric parameters -
	seeing, isoplanatic angle and effective altitude of turbulence. They are
	available via getval() and - errors - via geterr(). Also, the atmospheric
	time constant is computed  from provided \em desi and \em e2desi values.

	\b Seeing is computed as 
	\f[
		\beta = KS ~\lambda^{-1/5} M_0^{3/5}
	\f]
	where the zero-moment Mo of the turbulence is:
	\f[
		M_0 = \int C^2_n(h) dh
	\f]
	(integration goes from 0 for whole-atm. seeing and starting from \em 
	HBOUND for free-atm. seeing) and \em KS is the calibration constant.

	\b Isoplanatic angle is computed as 
	\f[
		\theta_0 = KP \lambda^{6/5} M_{5/3}^{-3/5} \sec^{8/5}(z)
	\f]
	where the 5/3-th moment of the turbulence profile is:
	\f[
		M_{3/5} = \int C^2_n(h) h^{5/3} dh
	\f]
	and \f$ KP=(2.91 \pi^2)^{-3/5} \f$ is the calibration constant.
	
	\b Second moment of the turbulence is related to the differential 
	astrometry precision and the operation of optical interferometers. Unlikely
	the isoplanatic angle, it is saved in files with no any calibration to any
	"angle".
	
	The \b free atmosphere \b effective \b altitude is computed as 
	\f[
		z_{eff}(free) = (M_{\alpha_1} / M_{\alpha_2})^{1/(\alpha_1-\alpha_2)}
	\f]
	where alpha_1 is 5/3 and alpha_2 is 1.
	
	The value of Heff in the \b whole atmophere is computed by definition:
	\f[
		z_{eff}(whole) = M_1 / M_0
	\f]
	but only if the shifted indices are available.
	
	The related to adaptive optics \b atmospheric \b time \b constant is
	computed from Differential Exposure Scintillation Index (DESI) \f$
	\sigma^2_{DE}\f$ as

	\f[
		\tau_{de} = KT \left[ \sigma^2_{DE} \right]^{-3/5}
	\f]
	where \em KT is the empirical calibration constant.

	\note
	During the conversion of integrals into atmospheric parameters, the
	integrals are checked to be positive. If at least one of integrals is
	invalid, the error nr::ERNOD is set which will block the writing of results
	by write() unless error is reset by nr::erreset() after avgint() and
	before write().
*/
void avgint(double desi, double e2desi, double zdist) ;

/**
	@brief	Write computed atmospheric parameters or restored Cn2 profile or
		the header line for one of them
	@param	f	FILE pointer
	@param	stamp	leading word in the output line
	@param	what	atm::ATMPAR: write atm. parameters, atm::CN2PROF: write 
		the Cn2 profile
	@param	header	Write the header line instead of data
	
	This utility writes the output results of atmospheric calculations on the
	disk. 
	
	For \em what==ATMPAR, the single line is written on disk, consisting
	of the average atmospheric parameters each followed by its error.
	Parameters are free seeing ["], seeing ["],	isoplanatic angle ["], second
	moment of turbulence [m^{4/3}],  and atmospheric time constant [ms].

	For \em what==CN2PROF, the Cn2-profile data line is written. It
	consists of 1) the number of adjusted Cn2 layers,  2) Chi-square 3) the
	getncn2() altitudes of layers and each followed by it strength in [m^-2/3].

	If \em header==true, the header line for respective \em what -selection is
	written instead of data. \em stamp must then contain the name of data which
	are normally written as a stamp when \em header==false.

*/
void write(FILE * f, const char * stamp, atmwhat what, bool header = false) ;

/**
	@brief	Get the average integral value
	@param	i	kind of the integral, see atm::ordint
	@return	array of averages

	Available after avgint() call.
*/
double getavgint(int i) ;

/**
	@brief	Get the \em squared error of the average integral
	@param	i	kind of the integral, see atm::ordint
	@return	array of error^2 of averages

	Available after avgint() call. Guaranteed to be non-negative.
*/
double geter2int(int i) ;

/*==========================================================================*/
/*                             Definitions                                  */
/*==========================================================================*/

/** Number of modes: normal and generalized, hense = 2 */
const int NMODE	= 2 ;

/** Minimal ratio of the singular value to the maximal one */
const double MINRELW = 1e-4 ;

/** Maximal number of orthonormal functions to use in h-powers restoration: (to
apply after MINRELW based rejection) */
const int MAXNBASE = 10 ;

/** Binary mask for using the shifted weights: "1"="use it".
	 Now we thus use all non-shifted and As, Bs and ABs weights. */
const char USESHFT[] = "1100100000" ;
//#define USESHFT "1111111111"

const double POWSEE = 0 ;		/**< Power of Cn2 moment for seeing */
const double POWISP = 5/3. ;	/**< Power of Cn2 moment for isoplanatic angle  */
const double POWEFF = 1.0 ;		/**< Power of Cn2 moment for Effective altitude */
const double POWISK = 2.0 ;		/**< Cn2 second moment power */

/** Height of the boundary layer [kilometers] to fit the Cn2-integral for 
free seeing (h^0 in integral is forced to resemble the AB-pair weight function
below this height) */
const double HBOUND = 1.0 ;

/** Number of power laws to fit: 2 kinds of 0-th, 5/3-th, 1-st, 2-nd */
const int NPOWER = 5 ;

/** Arrangement of the moment decomposition coefficients in the coef[] array: 
it's essential for calcint() that IWSEE goes in advance to IFSEE. */
enum poweridx {
IFSE,	/**< coef[Free seeing][]: */
IWSE,	/**< coef[Seeing][] */
IISP,	/**< coef[Isoplanatic angle][] */
IEFF,	/**< coef[Turbulence effective altitude][] */
IISK 	/**< coef[Isokinetic angle][] */
} ;

/** Calibration constant in seeing formula (AstL 45, p.399): 
\f$ seeing[''] = AT\_KS * \lambda_{eff}^{-1/5} M_0^{3/5} * 
\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/
const double KS = 1.73588e+07 ;

/** Calibration constant in isoplanatic angle formula (AstL 45, p.399)
\f$ \theta_0[''] = AT\_KP * (M_2 /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
\f$ AT\_KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
const double KP = 0.000756348 ;

/** Atmospheric time constant empirical calibration constant for DESI for 1,3ms
integrations in diameter=2cm apertures with -1km shift */
const double KT = 0.175 ;

/** Cn2-profile restoration weight matrix spacing: 1km (integer-like!)*/
const double DZCN2 = 1 ;

/** Formats of the records of average integral and their errors in the file. 
The order of strings follows the order of integrals specified in \em ordint.
Each format must be preceded with "\%s" to insert a delimiter specified in 
INTDLM.
*/
const char INTFMT[][8] = {"%s%6.3f","%s%6.3f","%s%6.3f","%s%8.2e","%s%5.0f",\
	"%s%5.0f","%s%5.2f"} ;

/** Delimiter of entries in the average data file */
const char INTDLM[] = " " ;

/** Altitude of layers record format with leading "\%s" for a delimiter */
const char ALTFMT[] =  "%s%2.0f" ;

/** Profile values format with leading "\%s" for a delimiter */
const char CN2FMT[] = "%s%8.2e" ;

} // of namespace at
#endif
