#include "star.h"
#include "names.h"
#include "atmos.hxx"
#include "scind.hxx"
#include "weif.hxx"
#include "nrutil.h"
#include "nr.h"

extern FILE* out_file;
extern double _tmZone;
extern QTime _programStartTime;

static QString str_message;

double toValidTimeInterval( double sec ) {
	double isec = floor( sec );
	double res = (double)( (int)isec % (int)SEC_IN_DAY ) +
			sec - isec;
	if( res < 0 ) res += SEC_IN_DAY;
	return( res );
}

double toDeg( double sec ) {
	return( sec / 3600.0 );
}

double toSec( int h, int m, int s ) {
	double sec = ( (double)h*60.0 + (double)m )*60.0 + (double)s; 
	return sec;
}

void toStrHour( double sec, QString& str ) {
	sec = toValidTimeInterval( sec );
	int _dd = (int)floor( sec / 3600.0 );
	int _dm = (int)floor( ( sec / 3600.0 - _dd ) * 60.0 );
	int _ds = (int)sec % 60;
	str.sprintf( " %02d  %02d  %02d  ", _dd,_dm,_ds );
}
 
void toStrHour12( double sec, QString& str ) {
	sec = toValidTimeInterval( sec );
	if( sec > SEC_IN_DAY / 2.0 ) sec -= SEC_IN_DAY; 
	char sign = ' ';
	if( sec < 0 ) sign = '-';
	sec = fabs( sec ); 
	int _dd = (int)floor( sec / 3600.0 );
	int _dm = (int)floor( ( sec / 3600.0 - _dd ) * 60.0 );
	int _ds = (int)sec % 60;
	str.sprintf( " %c%02d  %02d  %02d  ", sign,_dd,_dm,_ds );
}
 
void toStrDeg( double sec, QString& str ) {
	char* sign = "+";
	if( sec < 0 ) sign = "-";
	sec = fabs( sec );
	int _dd = (int)floor( sec / 3600.0 );
	int _dm = (int)floor( ( sec / 3600.0 - _dd ) * 60.0 );
	int _ds = (int)sec % 60;
	str.sprintf( " %s%02d  %02d  %02d  ", sign,_dd,_dm,_ds );
}


StarInfo::StarInfo( QString& _hr_,QString& _name_,double _ra_,double _dec_,
	QString& _magV_,QString& _sed_,QString& _spectrum_,QString& _duplicity_ ) {

	_hr = _hr_.copy();
	_name = _name_.copy();
	_ra = _ra_;
	_dec = _dec_;
	_decRad = _dec * RAD_IN_ASEC;
	_magV = _magV_.copy();
	_sed = _sed_.copy();
	_spectrum = _spectrum_.copy();
	_duplicity = _duplicity_.copy();
	_zenit = -1.0;
	_airMass = -1.0;
	_ha = 0.0;

}

StarInfo::~StarInfo() {
}

void StarInfo::update( double st, double sinLatitude,
			double cosLatitude ) {
	_ha  = toValidTimeInterval( st-_ra );
	double sinH = sinLatitude * sin( _decRad ) +
		cosLatitude * cos( _decRad ) * cos( _ha*RAD_IN_TSEC );
	if( (sinH >= -1) && (sinH <= 1) ) {
		_zenit = toDeg( SEC_IN_90 - asin( sinH ) / RAD_IN_ASEC );
	} else _zenit = -1.0;

	double secz1 = 1.0 / sinH;
	double secz2 = secz1 - 1.0;
	_airMass = secz1 - secz2 * ( 0.0018167 +
			secz2 * ( 0.002875 + 0.0008083 * secz2 ) );
}

QString& StarInfo::getRA( void ) {
	toStrHour( _ra, _ra_str );
	return( _ra_str );
}

QString& StarInfo::getHA( void ) {
	toStrHour12( _ha, _ha_str );
	return( _ha_str );
}

QString& StarInfo::getDec( void ) {
	toStrDeg( _dec, _dec_str );
	return( _dec_str );
}

QString& StarInfo::getAirMass( void ) {
	_airMass_str.setNum( _airMass,'f',2 );
	return( _airMass_str );
}

QString& StarInfo::getZenit( void ) {
	_zenit_str.setNum( _zenit,'f',2 );
	return( _zenit_str );
}

Star::Star( void ) {
	_nStars = 0;
	_currentStarNumber = 0;
	starList.resize( 100 );
}

Star::~Star() {
	debug( "Destructor of Star..." );
	int i;

	for( i=0; i<_nStars; i++ ) {
		delete starList[i];
	}
	debug( "End of destructor of Star" );
}

QString Star::composeWfFileName( QString& sed ) {
	QString fileName = sed.lower().prepend( _wfDirName )\
			.append( "." ).append( wfFileExt );
	return( fileName );
}

QString Star::composeSedFileName( QString& sed ) {
	QString	fileName = sed.lower().prepend( _sedDirName )\
			.append( "." ).append( sedFileExt );
	return( fileName );
}

void Star::loadConst( void ) {

	debug( "Star::loadConst   start..." );

	_longitude = 0;
	double latitude = 0;

	_tmZone = leafDataToDouble( cfgFile,"General",
		"Site", "TimeZone" );
	try {
		_longitude = leafDataToTimeSec( cfgFile,"General",
			"Site","Longitude",out_file );
		latitude = leafDataToDegSec( cfgFile,"General",
			"Site","Latitude",out_file );
	} catch( LoadError err ) {
		str_message = cfgFile;
		str_message.append( "\\General\\Site\\: " );
		str_message.append( err.getMessage() );
		showFatalError( str_message );
	}
	latitude *= RAD_IN_ASEC;
	_sinLatitude = sin( latitude );
	_cosLatitude = cos( latitude );

//	Optic:
	_innerDiam[0] = MM2CM * readDoubleConst( "Segmentator","Channel A","Inner",out_file );
	_outerDiam[0] = MM2CM * readDoubleConst( "Segmentator","Channel A","Outer",out_file );
	_innerDiam[1] = MM2CM * readDoubleConst( "Segmentator","Channel B","Inner",out_file );
	_outerDiam[1] = MM2CM * readDoubleConst( "Segmentator","Channel B","Outer",out_file );
	_innerDiam[2] = MM2CM * readDoubleConst( "Segmentator","Channel C","Inner",out_file );
	_outerDiam[2] = MM2CM * readDoubleConst( "Segmentator","Channel C","Outer",out_file );
	_innerDiam[3] = MM2CM * readDoubleConst( "Segmentator","Channel D","Inner",out_file );
	_outerDiam[3] = MM2CM * readDoubleConst( "Segmentator","Channel D","Outer",out_file );
	_magnification = readDoubleConst( "Optics","Common","Magnification",out_file );

//  WF:
	_minAltitude = readDoubleConst( "WeightFunction","MinAltitude",out_file );
	_maxAltitude = readDoubleConst( "WeightFunction","MaxAltitude",out_file );
	_gridStep = readDoubleConst( "WeightFunction","Step",out_file );
	_minGridStep = readDoubleConst( "WeightFunction","MinStep",out_file );

//	FILES & DIR-s:

	_responseFileName = *leafDataToStr( cfgFile,"General",
			"Inputs","Response",out_file );
	_responseFileName.prepend( dataPath );
	testFileName( _responseFileName,1 );

	_wfDirName = *leafDataToStr( cfgFile,"General",
		"Inputs","WeightDir" );
	if( _wfDirName.find( "/",_wfDirName.length()-1 ) == -1 ) {
		_wfDirName.append( "/" );
	}
	_wfDirName.prepend( dataPath );
	testDirName( _wfDirName,1 );

	_sedDirName = *leafDataToStr( cfgFile,"General",
			"Inputs","SpectraDir" );
	if( _sedDirName.find( "/",_sedDirName.length()-1 ) == -1 ) {
		_sedDirName.append( "/" );
	}
	_sedDirName.prepend( dataPath );
	testDirName( _sedDirName,1 );

	_starFileName = *leafDataToStr( cfgFile,"General",
			"Inputs","StarData" );
	_starFileName.prepend( dataPath );

	_programStartTime = QTime::currentTime()
		.addSecs( int(-1.0*_tmZone*SEC_IN_HOUR) );

	debug( "Star::loadConst   finish" );
}

void Star::init( void ) {
	try {
		loadStarData( 1 );
	} catch( LoadError err ) {
		str_message = err.getMessage();
		showFatalError( str_message );
	}

	atm::init( MAX_N_COUNTERS,_magnification, _innerDiam, _outerDiam,
		_minAltitude,_maxAltitude,_gridStep,_minGridStep,_responseFileName.data() );
	if( nr::erget() ) {
		str_message = nr::ermessage().c_str();
		nr::erreset();
		switch( showStrongWarning( str_message ) ) {
			case  QMessageBox::Abort : myQAppExit(1);
			case  QMessageBox::Ignore: break;
		}
	}

	QString wfFileName;
	QString sedFileName;
	for( int i=0; i<_nStars; i++ ) {
		sedFileName = composeSedFileName( starList[i]->getSED() );
		atm::loadsed( sedFileName.data() );
		if( nr::erget() ) {
			str_message = nr::ermessage().c_str();
			nr::erreset();
			showFatalError( str_message );
		} 
		wfFileName = composeWfFileName( starList[i]->getSED() );
		atm::checkwf( wfFileName.data() );
		if( nr::erget() ) {
			str_message = nr::ermessage().c_str();
			nr::erreset();
			switch( showStrongWarning( str_message,"Calculate" ) ) {
				case  QMessageBox::Abort : myQAppExit(1);
				case  QMessageBox::Ignore: break;
			}
			atm::calcwf( wfFileName.data(), wf::progress );
			if ( nr::erget() ) {
				debug( "%s",nr::ermessage().c_str() );
				str_message = nr::ermessage().c_str();
				nr::erreset();
				showFatalError( str_message );
			}
		}
	}
	updateAll();
	setCurrentStar( searchMinAirMass() );
}

void Star::setCurrentStar( int _number ) {
	_currentStarNumber = _number;
	QString sedFileName = composeSedFileName( starList[_number]->getSED() );
	atm::loadsed( sedFileName.data() );
	_wfFileName = composeWfFileName( starList[_number]->getSED() );
	atm::checkwf( _wfFileName.data() );
// Uzhe provereno po vsemu spisku, no malo li chto......
	if( nr::erget() ) {
		str_message = nr::ermessage().c_str();
		nr::erreset();
		showFatalError( str_message );
	}
	fprintf( out_file,"%s %s %s %s %s %s %s\n",
		getStamp( "O ",getCurrentUT() ),
		getHR().data(),
		getName().data(),
		getRA().data(),
		getDec().data(),
		getSED().data(),
		getZenit().data() );
}

QString Star::invalidFormat( QString& _fileName, int _nRow ) {
	str_message.sprintf( "\"%s\" Row %d: Invalid format!",
			_fileName.data(), _nRow );
	return( str_message );
}

QString Star::notFound( QString& _fileName ) {
	str_message.sprintf( "\"%s\" not found!",
			_fileName.data() );
	return( str_message );
}

/* Calculation of the Julian date */

long Star::julian( QDate utDate ) {

	int month = utDate.month();
	int day = utDate.day();
	int year = utDate.year();

	long result = (3057*(long)month)/100 + 
				(365*(long)year) + 1721028L;
	result = result + (long)(year/4 - year/100 + year/400 + day);
	if( month > 2 ) return( result);
	int j = 2;
	if( year%4 == 0 )	j -= 1;
	if( year%100 == 0 )	j -= 1;
	if( year%400 == 0 )	j += 1;
	result += j;
	return( result );
}

/* Calculation of siderial time in the Greewhich midnigth */

double Star::calcS0Time( QDate utDate ) {
	double c3		= 1.0775925925925926E-06;
	double c2		= 2.7379093507870371E-03;
	double c1		= 0.2790572732638889;

	double d = (double)julian( utDate ) - (double)2451545.5;
	double t2 = d/(double)36525.0;
	double result = c1 + c2*d + c3*t2*t2;
	result = result - floor( result );
	return( result );
}

/* Calculation of Current Local Sideral Time */

double Star::calcLocalST( void ) {
	QDateTime utDateTime = QDateTime::currentDateTime();
	utDateTime = utDateTime.addSecs( int(-1.0*_tmZone*SEC_IN_HOUR) ); // to UT

	double sec = toSec( utDateTime.time().hour(),
			utDateTime.time().minute(),utDateTime.time().second() );
	double st = calcS0Time( utDateTime.date() ) 
			+ MU*(double)sec / SEC_IN_DAY;
	st = SEC_IN_DAY*( st - floor( st ) );
	double stLoc = st + _longitude;
	return( stLoc );
}

QString Star::invalidDec( QString& _fileName, int _nRow ) {
	str_message.sprintf( "\"%s\" Row %d: Invalid Dec!",
			_fileName.data(), _nRow );
	return( str_message );
}

QString Star::invalidRA( QString& _fileName, int _nRow ) {
	str_message.sprintf( "\"%s\" Row %d: Invalid RA!",
			_fileName.data(), _nRow );
	return( str_message );
}

QString Star::emptyFile( QString& _fileName ) {
	str_message.sprintf( "\"%s\" is empty!",
			_fileName.data() );
	return( str_message );
}

int Star::loadStarData( int _dialog ) {
	if( testFileName( _starFileName, _dialog ) != 1 ) return(0);
	int loadError = 0;
	bool ok;
	int _h,_m,_s,_sign,_dd,_dm,_ds;

	QFile file( _starFileName );
	if( !file.open( IO_ReadOnly ) ) throw LoadError( notFound( _starFileName ) );
	QTextStream textStream( &file );
	_nStars = 0;
	int _nRow = 1;
	QString _magV_,_hr_,_name_,_sed_,_spectrum_,_duplicity_;
	double _ra_,_dec_;
	while( !textStream.eof() && !loadError ) {
		str_message = textStream.readLine();
		if( (str_message[0] != '#') && (!str_message.isEmpty()) ) {

			str_message.replace( QRegExp(","),QString(" ") );

			_hr_ = str_message.left( 4 );
			_name_ = str_message.mid( 4,9 ) + ' ';

			_h = str_message.mid( 13,2 ).toInt( &ok );
			if( !ok ) {
				throw LoadError( invalidFormat( _starFileName,_nRow ) );
			}

			_m = str_message.mid( 16,2 ).toInt( &ok );
			if( !ok ) throw LoadError( invalidFormat( _starFileName,_nRow ) );
			_s = str_message.mid( 19,2 ).toInt( &ok );
			if( !ok ) throw LoadError( invalidFormat( _starFileName,_nRow ) );

			if( str_message[22] == '-' ) _sign = -1;
			else _sign = 1;

			_dd = str_message.mid( 24,2 ).toInt( &ok );
			if( !ok ) throw LoadError( invalidFormat( _starFileName,_nRow ) );
			_dm = str_message.mid( 27,2 ).toInt( &ok );
			if( !ok ) throw LoadError( invalidFormat( _starFileName,_nRow ) );
			_ds = str_message.mid( 30,2 ).toInt( &ok );
			if( !ok ) throw LoadError( invalidFormat( _starFileName,_nRow ) );

			_ra_ = toSec( _h,_m,_s );
			if( (_ra_ < 0) || (_ra_ > SEC_IN_DAY) ) throw LoadError( invalidRA( _starFileName,_nRow ) );
			_dec_ = _sign * toSec( _dd,_dm,_ds );
			if( fabs(_dec_) > SEC_IN_90 ) throw LoadError( invalidDec( _starFileName,_nRow ) );

			_magV_ = str_message.mid( 33,5 );
			_sed_ = str_message.mid( 44,5 ).stripWhiteSpace();
			_spectrum_ = str_message.mid( 48,7 ).stripWhiteSpace();
			
			_duplicity_ = str_message.right( str_message.length()-65 ).stripWhiteSpace();

			if( _nStars >= (int)starList.size() ) {
				starList.resize( starList.size()+10 );
			}
			starList[_nStars] = new StarInfo( _hr_,_name_,_ra_,
					_dec_,_magV_,_sed_,_spectrum_,_duplicity_ );
			_nStars++;
		}
		_nRow++;
	}
	file.close();
	if( _nStars <= 0 ) throw LoadError( emptyFile( _starFileName ) );
	return(1);
}

int Star::searchMinAirMass( void ) {
	int n = 0;
	double minAirMass = 10.0;
	for( int i=0; i<_nStars; i++ ) {
		double am = starList[i]->getAirMass().toDouble();
		if( (am >= 1.0) && (am < minAirMass) ) {
			n = i;
			minAirMass = am;
		}
	}
	return( n );
}

void Star::updateAll() {
	int i;
	double st = calcLocalST();
	for( i=0; i<_nStars; i++ ) {
		starList[i]->update( st,_sinLatitude,_cosLatitude );
	}
}

void Star::updateCurrent() {
	double st = calcLocalST();
	starList[_currentStarNumber]->
			update( st,_sinLatitude,_cosLatitude );
}

QString& Star::getSpectrum( void ) {
	return( starList[_currentStarNumber]->getSpectrum() );
}

QString& Star::getSED( void ) {
	return( starList[_currentStarNumber]->getSED() );
}

QString& Star::getName( void ) {
	return( starList[_currentStarNumber]->getName() );
}

QString& Star::getHR( void ) {
	return( starList[_currentStarNumber]->getHR() );
}

QString& Star::getMagV( void ) {
	return( starList[_currentStarNumber]->getMagV() );
}

QString& Star::getDuplicity( void ) {
	return( starList[_currentStarNumber]->getDuplicity() );
}

QString& Star::getRA( void ) {
	return( starList[_currentStarNumber]->getRA() );
}

QString& Star::getHA( void ) {
	return( starList[_currentStarNumber]->getHA() );
}

QString& Star::getDec( void ) {
	return( starList[_currentStarNumber]->getDec() );
}

QString& Star::getZenit( void ) {
	return( starList[_currentStarNumber]->getZenit() );
}

double Star::getZenitDist( void ) {
	return( starList[_currentStarNumber]->getZenitDist() );
}

QString& Star::getAirMass( void ) {
	return( starList[_currentStarNumber]->getAirMass() );
}

