#ifndef STAR_H
#define STAR_H

#include <math.h>
#include <qstring.h>
#include <qarray.h>
#include <qdatetime.h>
#include "jkerror.h"
#include "devplus.h"
#include "tools.h"

#define	MU				(1.0027379093)
#define	RAD_IN_TSEC		(0.7272220521664304e-4)
#define	RAD_IN_ASEC		(0.484813681109536e-5)
#define	PIM2			(6.28318530717958647712)
#define	SEC_IN_DAY		86400.0		// 24 hour
#define	SEC_IN_90		324000.0	// 90 deg

double toValidTimeInterval( double );
double toDeg( double );
double toSec( int,int,int );
void toStrHour( double, QString& );		// 0-24h
void toStrHour12( double, QString& );	// +- 12h
void toStrDeg( double, QString& );

class StarInfo {
private:
	QString	_hr;
	QString	_name;

	double	_ra;		// ra in tsec
	double	_dec;		// declination in asec
	double	_decRad;	// declination in radians
	double	_ha;		// hour angle in tsec
	double	_zenit;		// deg
	double	_airMass;
	QString	_ra_str;
	QString	_dec_str;
	QString	_ha_str;
	QString	_zenit_str;
	QString	_airMass_str;

	QString	_magV;
	QString	_sed;
	QString _spectrum;
	QString	_duplicity;

public:

	StarInfo( QString&, QString&, double, double,
			QString&, QString&, QString&, QString& );
	~StarInfo();
	void	update( double, double, double );

	QString& getHR( void ) { return _hr; }
	QString& getName( void ) { return _name; }
	QString& getMagV( void ) { return _magV; }
	QString& getSpectrum( void ) { return _spectrum; }
	QString& getSED( void ) { return _sed; }
	QString& getDuplicity( void ) { return _duplicity; }
	QString& getRA( void );
	QString& getHA( void );
	QString& getDec( void );
	QString& getZenit( void );
	double	getZenitDist( void ) { return _zenit; }
	QString& getAirMass( void );
};

class Star {

private:
	QArray<StarInfo*>	starList;
	int					_nStars;
	int					_currentStarNumber;

// Files & Dirs:
	QString				_sedDirName;
	QString				_wfDirName;
	QString				_wfFileName;
	QString				_responseFileName;
	QString				_starFileName;

// Optic:
	double				_magnification;
	double				_innerDiam[MAX_N_COUNTERS];
	double				_outerDiam[MAX_N_COUNTERS];

// Weight Functions:
	double _minAltitude, _maxAltitude, _gridStep, _minGridStep;

// Site:
	double				_longitude;	// h_sec
	double				_sinLatitude;
	double				_cosLatitude;

public:
	Star( void );
	~Star();

	QString		composeWfFileName( QString& );
	QString		composeSedFileName( QString& );

	void		loadConst( void );
	void		init( void );

	QString		notFound( QString& );
	QString		invalidFormat( QString&, int );
	QString		invalidRA( QString&, int );
	QString		invalidDec( QString&, int );
	QString		emptyFile( QString& );

	int			loadStarData( int );
	long		julian( QDate );
	double		calcS0Time( QDate );
	double		calcLocalST( void );

	int			searchMinAirMass( void );
	void		updateAll( void );
	void		updateCurrent( void );
	QString&	getHR( void );
	QString&	getName( void );
	QString&	getSpectrum( void );
	QString&	getSED( void );
	QString&	getWFName( void ) { return _wfFileName; }
	QString&	getDuplicity( void );
	QString&	getMagV( void );
	QString&	getRA( void );
	QString&	getHA( void );
	QString&	getDec( void );
	QString&	getZenit( void );
	double		getZenitDist( void );
	QString&	getAirMass( void );
	QArray<StarInfo*>&	getStarList( void ) { return starList; }
	int			getNStars( void ) { return _nStars; }
	void		setCurrentStar( int );
};

#endif
