#include "widgets.h"
#include "star.h"
#include "devplus.h"
#include "tools.h"
#include "jkexport.h"

extern DevManager	devmanager;
extern Star		star;
extern QMutex		stateExportMutex;
extern ExportTable	table;

static const char* starText			= "Observed Star";
static const char* haText			= "Hour Angle";
static const char* raText			= "RA";
static const char* decText			= "Declination";
static const char* magText			= "V-magnitude";
static const char* spText			= "Spectrum";
static const char* duplText			= "Duplicity";

static const int SEPARATOR = 5;

int getHeight( int fontHeight ) {
	return fontHeight*5 / 3;
}
QJKMainView::QJKMainView( QWidget* parent=0,const char* name=0 )
				:QWidget( parent, name ) {
	debug( "QJKMainView: start Constructor" );
	table = new QJKTable( this, "MainTable", MAX_ROW, MAX_COL );
	CHECK_PTR( table );

	textView = new QLineEdit( this, "MainTextView" );
	CHECK_PTR( textView );
	textView->setReadOnly( true );
	textView->setAlignment( AlignHCenter );

	setChildrenGeometry();
	debug( "QJKMainView: end Constructor" );
}

QJKMainView::~QJKMainView() {
}
                        
void QJKMainView::fillTable( void ) {
	table->fillTable();
}

void QJKMainView::showTextLine( void ) {
	debug( "TextView setText: %s",(ExportText::text).data() );
	textView->setText( ExportText::text );
}

void QJKMainView::setChildrenGeometry( void ) {
	int _height = getHeight( fontMetrics().height() );
	debug( "%d %d %d",
		width(),height(),_height );
	table->setGeometry( 0,0,width(),
		height() - _height - SEPARATOR );
	textView->setGeometry( 0,table->y()+table->height()+SEPARATOR,
			width(), _height );
	table->setFrameStyle( QFrame::Sunken );
}

void QJKMainView::resizeEvent( QResizeEvent* e ) {
	debug( "QJKMainView: ResizeEvent!" );
	setChildrenGeometry();
}

void QJKMainView::fitToFont( void ) {
	table->fitToFont();
}

QJKTable::QJKTable( QWidget* parent=0, const char* name=0,
			int numRows=MAX_ROW, int numCols=MAX_COL,
			bool grid=false )
				:QTableView( parent, name ) {
	setFocusPolicy( NoFocus );
	setBackgroundMode( PaletteBase );
	setNumCols( numCols );
	setNumRows( numRows );

	setCellGeometry();

	setTableFlags(
		Tbl_clipCellPainting	// avoid drawing outside cell
		);

	contents = new QString[numRows * numCols]; //make room for contents


//	isColumnSizeChanged = false;
	showGrid = grid;
}

QJKTable::~QJKTable() {
	debug( "destructor of Table" );
	delete[] contents;
}

void QJKTable::setCellContent( int row, int col, const QString& text ) {
	contents[ indexOf( row,col )] = text;
	updateCell( row,col );	// show new content
}

void QJKTable::paintCell( QPainter* p, int row, int col ) {
	int w = cellWidth( col );	// width of cell in pixels
	int h = cellHeight( row );	// height of cell in pixels

	if( showGrid ) {
		int x2 = w - 1;
		int y2 = h - 1;
	// Draw our part of cell frame.
		p->drawLine( x2,0,x2,y2 );	// draw vertical line on right
		p->drawLine( 0, y2, x2, y2 );	// raw horiz. line at bottom
	}

// Draw cell content (text)	
	p->drawText( 0,0,w,h,AlignCenter,contents[indexOf(row,col)] );

}

int QJKTable::indexOf( int row, int col ) const {
	return( row * numCols() + col );
}

void QJKTable::setCellGeometry( void ) {
	int cellWidth = (int)( (double)width() / ((double)numCols()+0.25) );
	int cellHeight = fontMetrics().height();
	setCellWidth( cellWidth  );
	setCellHeight( cellHeight );
}

void QJKTable::resizeEvent( QResizeEvent* e ) {
	setCellGeometry();
}

void QJKTable::fillTable( void  ) {

	int startRow = table.startRow;
	int endRow = table.startRow + table.nRows;
	int startCol = table.startCol;
	int endCol = table.startCol + table.nCols;
	if( startRow < 0 ) startRow = 0;
	if( startCol < 0 ) startCol = 0;
	if( endRow > numRows() ) endRow = numRows();
	if( endCol > numCols()) endCol = numCols();

	debug( "fillTable: %d %d %d %d",startRow,endRow,
			startCol,endCol );
	if( table.clear ) {
		table.clear = false;
		debug( "Clearing!" );
		for( int row=0; row<numRows(); row++ ) {
			for( int col=0; col<numCols(); col++ ) {
				setCellContent( row,col,"" );
			}
		}
	}
	for( int row=startRow; row<endRow; row++ ) {
		for( int col=startCol; col<endCol; col++ ) {
			setCellContent( row,col,table.getData( row,col ) );
		}
	}
}

void QJKTable::fitToFont( void ) {
	setCellGeometry();
}

QJKTextView::QJKTextView( QWidget* parent=0,const char* name=0 )
				:QTextView( parent, name ) {
	setVScrollBarMode( QScrollView::AlwaysOff );
	setHScrollBarMode( QScrollView::AlwaysOff );
}

QJKTextView::~QJKTextView() {
	debug( "destructor of TextView" );
}

void QJKTextView::addLine( const QString& _text ) {
	text = _text;
	setText( text );
}

QJKPushButton::QJKPushButton( const char* text, 
			QWidget* parent=0,const char* name=0 )
		: QPushButton( text, parent, name ) {
}

QJKPushButton::~QJKPushButton() {
}

void QJKPushButton::focusInEvent( QFocusEvent* e ) {
	setDefault( true );
	QPushButton::focusInEvent( e );
}

void QJKPushButton::focusOutEvent( QFocusEvent* e ) {
	setDefault( false );
	QPushButton::focusOutEvent( e );
}

DigitalClock::DigitalClock( QWidget *parent, const char *name )
				: QLCDNumber( parent, name ) {
	setNumDigits(8);
	setSegmentStyle( Flat );
	showTime();
}

DigitalClock::~DigitalClock() {
}

void DigitalClock::showTime() {
	QString s = QTime::currentTime().toString().left(8);
	if( s[0] == '0' ) s[0] = ' ';
	display( s );   // set LCD number/text
}

TalkDialog::TalkDialog( void )	
		: QDialog( 0,"TalkDialog",true /*modal*/ ) {
	setCaption( "Talk" );

	int border = fontMetrics().height();
	layout = new QVBoxLayout( this,border );
	CHECK_PTR(layout);

	label = new QLabel( "Talk",this );
	CHECK_PTR( label );
	label->setFixedSize( label->sizeHint() );
	layout->addWidget( label );

	combo = new QComboBox( true,this );
	CHECK_PTR( combo );
	int _width = fontMetrics().width( "wwwwwwwwwww" );
	int _height = getHeight( fontMetrics().height() );
	combo->setFixedSize( _width, _height );
	combo->setFocus();
	connect( combo,SIGNAL( activated( const QString& ) ),
		this,SLOT( slotTalk( const QString& ) ) );

	layout->addWidget( combo );

	edit = new QMultiLineEdit( this );
	CHECK_PTR( edit );
	edit->setFixedSize( _width,3*_height );
	edit->setReadOnly( true );
	edit->setFocusPolicy( NoFocus );
	layout->addWidget( edit );
	
	closebutton = new QJKPushButton( "Close",this );
	CHECK_PTR(closebutton);
	closebutton->setMinimumSize( closebutton->sizeHint() );
	connect( closebutton, SIGNAL( clicked() ), 
				this, SLOT( accept() ) );

	layout->addWidget( closebutton );
}

TalkDialog::~TalkDialog() {
	debug( "TalkDialog destructor..." );
	delete layout;
	delete label;
	delete combo;
	delete edit;
	delete closebutton;
}

void TalkDialog::slotTalk( const QString& text ) {
	const char* resp;
	resp = talk( text );
	debug( "slotTalk: response: %s", resp );
	edit->setText( resp );
}

AdjustDialog::AdjustDialog( void )	
		: QDialog( 0,"AdjustDialog",true /*modal*/ ) {

	lastAperture = devmanager.currentAperture();
	setCaption( "Aperture adjust" );

	int border = fontMetrics().height();

	layout = new QGridLayout( this,5,3,border );
	CHECK_PTR(layout);

	label = new QLabel( "Aperture adjust:",this );
	CHECK_PTR( label );
	layout->addMultiCellWidget( label,0,0,0,2,AlignHCenter );

	combo_apert = new QComboBox( false,this );
	CHECK_PTR( combo_apert );

	int i;
	int ns = devmanager.getNStops();
	for( i=0; i<ns; i++ ) {
		combo_apert->insertItem( devmanager.getStopName(i) );
	}	
//	connect( &devmanager, SIGNAL( signalAperture(int) ),
//		this, SLOT( setCurrentItem(int) ) );
	connect( combo_apert,SIGNAL( activated( const QString& ) ),
		this,SLOT( setAperture( const QString& ) ) );
	layout->addMultiCellWidget( combo_apert,1,1,1,2,AlignHCenter );
	label_apert = new QLabel( "Aperture:",this );
	CHECK_PTR( label_apert );
	layout->addWidget( label_apert,1,0,AlignLeft );

	label_correct = new QLabel( "Correction:",this );
	CHECK_PTR( label_correct );
	layout->addWidget( label_correct,2,0,AlignLeft );
	result_window = new QLineEdit( this );
	CHECK_PTR( result_window );
	result_window->setReadOnly( true );
	result_window->setAlignment( AlignHCenter );
	int _width = fontMetrics().width( "wwwwwww" );
	int _height = getHeight( fontMetrics().height() );
	result_window->setFixedSize( _width, _height );
	QString str;
	str.setNum( devmanager.getStopCorrectNew() );
	result_window->setText( str );
	layout->addMultiCellWidget( result_window,2,2,1,2,AlignHCenter );

	move_plus = new QJKPushButton( "+",this );
	CHECK_PTR(move_plus);
	move_plus->setAccel('+');
	connect( move_plus, SIGNAL( clicked() ), 
		this, SLOT( movePlus() ) );
	layout->addWidget( move_plus,3,2 );

	move_minus = new QJKPushButton( "-",this );
	CHECK_PTR(move_minus);
	move_minus->setAccel('-');
	connect( move_minus, SIGNAL( clicked() ), 
		this, SLOT( moveMinus() ) );
	layout->addWidget( move_minus,3,1 );

	label_move = new QLabel( "Move:",this );
	CHECK_PTR( label_move );
	layout->addWidget( label_move,3,0,AlignLeft );

	acceptbutton = new QJKPushButton( "Accept",this );
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize( acceptbutton->sizeHint() );
	connect( acceptbutton, SIGNAL( clicked() ), 
				this, SLOT( accept() ) );
	layout->addWidget( acceptbutton,4,0 );

	cancelbutton = new QJKPushButton( "Cancel",this );
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize( cancelbutton->sizeHint() );
	connect( cancelbutton, SIGNAL( clicked() ), 
				this, SLOT( reject() ) );
	layout->addMultiCellWidget( cancelbutton,4,4,1,2,AlignHCenter );
}

void AdjustDialog::update( void ) {
//	setCurrentItem( devmanager.currentAperture() );
	devmanager.setAperture( lastAperture,1 );
	setCurrentItem( lastAperture );
	debug( "AdjusDialog::update setAperture %d",lastAperture );
}

void AdjustDialog::moveMinus( void ) {
	QString str;

	devmanager.moveMinus();
	str.setNum( devmanager.getStopCorrectNew() );
	result_window->setText( str );
}

void AdjustDialog::movePlus( void ) {
	QString str;

	devmanager.movePlus();
	str.setNum( devmanager.getStopCorrectNew() );
	result_window->setText( str );
}

void AdjustDialog::setCurrentItem( int _item ) {
	QString str;
	debug( "setCurrentItem item=%d",_item );

	if( (_item >= 0) && (_item < combo_apert->count()) ) {
		if( combo_apert ) combo_apert->setCurrentItem( _item );
		str.setNum( devmanager.getStopCorrectNew() );
		debug( "str=%s",str.data() );
		if( result_window ) result_window->setText( str );
	}
	debug( "return from setCurrentItem" );
}

void AdjustDialog::setAperture( const QString& _apertName ) {
	devmanager.setAperture( _apertName,1 );
	lastAperture = devmanager.currentAperture();
	setCurrentItem( lastAperture );
}

AdjustDialog::~AdjustDialog() {
	debug( "AdjustDialog destructor..." );
	delete layout;
	delete label;
	delete label_apert;
	delete label_move;
	delete label_correct;
	delete result_window;
	delete combo_apert;
	delete acceptbutton;
	delete cancelbutton;
	delete move_plus;
	delete move_minus;
}

ShowStarDialog::ShowStarDialog( int _showHA_ )
		: QDialog( 0,"ShowStarDialog",true ) {
	_showHA = _showHA_;
	setCaption( "Show star" );

	int border = fontMetrics().height();

	layout = new QGridLayout( this,3,6,border );
	CHECK_PTR(layout);

	labelName = new QLabel( star.getName(),this );
	CHECK_PTR( labelName );
	QToolTip::add( labelName,starText );
	layout->addMultiCellWidget( labelName,0,0,0,5,AlignHCenter );
	
	if( _showHA ) {
		star.updateCurrent();
		labelHA = new QLabel( star.getHA(),this );
		CHECK_PTR( labelHA );
		QToolTip::add( labelHA,haText );
	} else {
		labelHA = new QLabel( star.getRA(),this );
		CHECK_PTR( labelHA );
		QToolTip::add( labelHA,raText );
	}
	labelHA->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	labelHA->setAlignment( AlignCenter );
	layout->addWidget( labelHA,1,0 );

	labelDec = new QLabel( star.getDec(),this );
	CHECK_PTR( labelDec );
	QToolTip::add( labelDec,decText );
	labelDec->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	labelDec->setAlignment( AlignCenter );
	layout->addWidget( labelDec,1,1 );

	labelMagV = new QLabel( star.getMagV(),this );
	CHECK_PTR( labelMagV );
	QToolTip::add( labelMagV,magText );
	labelMagV->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	labelMagV->setAlignment( AlignCenter );
	layout->addWidget( labelMagV,1,2 );

	labelSp = new QLabel( star.getSpectrum(),this );
	CHECK_PTR( labelSp );
	QToolTip::add( labelSp,spText );
	labelSp->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	labelSp->setAlignment( AlignCenter );
	layout->addWidget( labelSp,1,3 );

	labelDupl = new QLabel( star.getDuplicity(),this );
	CHECK_PTR( labelDupl);
	QToolTip::add( labelDupl,duplText );
	labelDupl->setFrameStyle( QFrame::Panel | QFrame::Sunken );
	labelDupl->setAlignment( AlignCenter );
	layout->addMultiCellWidget( labelDupl,1,1,4,5,AlignHCenter );

	closeButton = new QJKPushButton( "Close",this );
	CHECK_PTR( closeButton );
	layout->addWidget( closeButton,2,3 );
	connect( closeButton, SIGNAL( clicked() ), 
				this, SLOT( accept() ) );
}

ShowStarDialog::~ShowStarDialog() {
	delete layout;
	delete labelName;
	delete labelHA;
	delete labelDec;
	delete labelMagV;
	delete labelSp;
	delete labelDupl;
	delete closeButton;
}

void ShowStarDialog::update( int _showHA_ ) {
	_showHA = _showHA_;
	labelName->setText( star.getName() );
	QToolTip::remove( labelHA );
	if( _showHA ) {
		star.updateCurrent();
		labelHA->setText( star.getHA() );
		QToolTip::add( labelHA,haText );
	} else {
		labelHA->setText( star.getRA() );
		QToolTip::add( labelHA,raText );
	}
	labelDec->setText( star.getDec() );
	labelMagV->setText( star.getMagV() );
	labelSp->setText( star.getSpectrum() );
	labelDupl->setText( star.getDuplicity() );
}

void ShowStarDialog::showHA( void ) {
	star.updateCurrent();
	labelHA->setText( star.getHA() );
}

SelectStarDialog::SelectStarDialog( void )	
		: QDialog( 0,"SelectStarDialog",true /*modal*/ ) {

	setCaption( "Select star" );

	int border = fontMetrics().height();

	layout = new QGridLayout( this,7,2,border );
	CHECK_PTR(layout);

	label = new QLabel( "Select star:",this );
	CHECK_PTR( label );
	label->setFixedSize( label->sizeHint() );
	layout->addMultiCellWidget( label,0,0,0,1,AlignHCenter );

	viewStarList = new ViewStarList( &star, this );
	viewStarList->setFixedSize( viewStarList->sizeHint() );
	layout->addMultiCellWidget( viewStarList,1,5,0,1,AlignHCenter );

	selectbutton = new QJKPushButton( "Select",this );
	CHECK_PTR(selectbutton);
	selectbutton->setFixedSize( selectbutton->sizeHint() );
	layout->addWidget( selectbutton,6,0 );
	connect( selectbutton, SIGNAL( clicked() ), 
				this, SLOT( select() ) );
	connect( viewStarList, SIGNAL( doubleClicked(QListViewItem*) ),
				this, SLOT( select(QListViewItem*) ) );

	cancelbutton = new QJKPushButton( "Cancel",this );
	CHECK_PTR(cancelbutton);
	layout->addWidget( cancelbutton,6,1 );
	cancelbutton->setFixedSize( cancelbutton->sizeHint() );
	connect( cancelbutton, SIGNAL( clicked() ), 
				this, SLOT( reject() ) );
}

void SelectStarDialog::update( void ) {
	viewStarList->update( &star );
}

void SelectStarDialog::select( void ) {
	int number = viewStarList->getCurrentStarNumber();
	star.setCurrentStar( number );
	accept();
}

void SelectStarDialog::select( QListViewItem* it ) {
	int number = viewStarList->getCurrentStarNumber();
	star.setCurrentStar( number );
	accept();
}

SelectStarDialog::~SelectStarDialog() {
	debug( "SelectStarDialog destructor..." );
	delete label;
	delete layout;
	delete viewStarList;
	delete selectbutton;
	delete cancelbutton;
}

IllumDialog::IllumDialog( void )	
		: QDialog( 0,"IllumDialog",true /*modal*/ ) {
	setCaption( "Illumination level" );

	int border = fontMetrics().height();

	layout = new QGridLayout( this,4,2,border );
	CHECK_PTR(layout);

	label = new QLabel( "Illumination level:",this );
	CHECK_PTR( label );
	label->setFixedSize( label->sizeHint() );
	layout->addMultiCellWidget( label,0,0,0,1,AlignHCenter );

	bright_window = new QLineEdit( this );
	bright_window->setReadOnly( true );
	bright_window->setAlignment( AlignHCenter );
	CHECK_PTR( bright_window );
	int _width = fontMetrics().width( "wwwwwww" );
	int _height = getHeight( fontMetrics().height() );
	bright_window->setFixedSize( _width, _height );
//	setBrightness();
//	connect( &devmanager, SIGNAL( signalBrightness(int) ), 
//		this, SLOT( slotSetBrightness(int) ) );
	layout->addMultiCellWidget( bright_window,1,1,0,1,AlignHCenter );

	bright_plus = new QJKPushButton( "+",this );
	CHECK_PTR(bright_plus);
	bright_plus->setAccel('+');
	bright_plus->setFixedSize( bright_plus->sizeHint() );
	connect( bright_plus, SIGNAL( clicked() ), 
		this, SLOT( setBrightnessUp() ) );
//		&devmanager, SLOT( setBrightnessUp() ) );
	layout->addWidget( bright_plus,2,1 );

	bright_minus = new QJKPushButton( "-",this );
	CHECK_PTR(bright_minus);
	bright_minus->setAccel('-');
	bright_minus->setFixedSize( bright_minus->sizeHint() );
	connect( bright_minus, SIGNAL( clicked() ), 
		this, SLOT( setBrightnessDown() ) );
//		&devmanager, SLOT( setBrightnessDown() ) );
	layout->addWidget( bright_minus,2,0 );

	closebutton = new QJKPushButton( "Close",this );
	CHECK_PTR(closebutton);
	closebutton->setMinimumSize( closebutton->sizeHint() );
	connect( closebutton, SIGNAL( clicked() ), 
				this, SLOT( accept() ) );
	layout->addMultiCellWidget( closebutton,3,3,0,1,AlignHCenter );
}

void IllumDialog::update( void ) {
	setBrightness();
}

void IllumDialog::setBrightnessUp( void ) {
	devmanager.setBrightnessUp();
	setBrightness();
}

void IllumDialog::setBrightnessDown( void ) {
	devmanager.setBrightnessDown();
	setBrightness();
}

void IllumDialog::setBrightness( void ) {
	QString str;

	str.sprintf( " %.2f",
		(double)devmanager.getCurrentBrightness() / 15.0 );
	bright_window->setText( str );
}

IllumDialog::~IllumDialog() {
	debug( "IllumDialog destructor..." );
	delete layout;
	delete label;
	delete bright_window;
	delete bright_plus;
	delete bright_minus;
	delete closebutton;
}

BackCorrectionDialog::BackCorrectionDialog( void )
			: QDialog( 0,"Background Correction",true ) {
	setCaption( "Background Correction" );
	int border = fontMetrics().height();
	layout = new QGridLayout( this,MAX_N_COUNTERS+3,5,border );
	CHECK_PTR(layout);

	label = new QLabel( "Background Correction",this );
	CHECK_PTR( label );
	layout->addMultiCellWidget( label,0,0,0,3,AlignHCenter );

	label_cf = new QLabel( "Center/Side ratio",this );
	CHECK_PTR( label_cf );
	layout->addMultiCellWidget( label_cf,1,1,0,3,AlignHCenter );

	label_l = new QLabel( "Left",this );
	CHECK_PTR( label_l );
	layout->addWidget( label_l,2,1,AlignHCenter );

	label_r = new QLabel( "Right",this );
	CHECK_PTR( label_r );
	layout->addWidget( label_r,2,2,AlignHCenter );

	label_av = new QLabel( "Average",this );
	CHECK_PTR( label_av );
	layout->addWidget( label_av,2,3,AlignHCenter );

	label_a = new QLabel( "Chan.A",this );
	CHECK_PTR( label_a );
	layout->addWidget( label_a,3,0,AlignHCenter );

	label_b = new QLabel( "Chan.B",this );
	CHECK_PTR( label_b );
	layout->addWidget( label_b,4,0,AlignHCenter );

	label_c = new QLabel( "Chan.C",this );
	CHECK_PTR( label_c );
	layout->addWidget( label_c,5,0,AlignHCenter );

	label_d = new QLabel( "Chan.D",this );
	CHECK_PTR( label_d );
	layout->addWidget( label_d,6,0,AlignHCenter );

	int _width = fontMetrics().width( "wwww" );
	int _height = getHeight( fontMetrics().height() );
	int i;

	for( i=0; i<MAX_N_COUNTERS; i++ ) {
		edit_l[i] = new QLineEdit( this );
		CHECK_PTR( edit_l[i] );
		edit_l[i]->setReadOnly( true );
		edit_l[i]->setAlignment( AlignHCenter );
		edit_l[i]->setFixedSize( _width, _height );
		edit_l[i]->setFocusPolicy( QWidget::NoFocus );
		layout->addWidget( edit_l[i],3+i,1 );

		edit_r[i] = new QLineEdit( this );
		CHECK_PTR( edit_r[i] );
		edit_r[i]->setReadOnly( true );
		edit_r[i]->setAlignment( AlignHCenter );
		edit_r[i]->setFixedSize( _width, _height );
		edit_r[i]->setFocusPolicy( QWidget::NoFocus );
		layout->addWidget( edit_r[i],3+i,2 );

		edit_av[i] = new QLineEdit( this );
		CHECK_PTR( edit_av[i] );
		edit_av[i]->setReadOnly( true );
		edit_av[i]->setAlignment( AlignHCenter );
		edit_av[i]->setFixedSize( _width, _height );
		edit_av[i]->setFocusPolicy( QWidget::NoFocus );
		layout->addWidget( edit_av[i],3+i,3 );
	}

	repeatbutton = new QJKPushButton( "&Measure",this );
	CHECK_PTR(repeatbutton);
	repeatbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addMultiCellWidget( repeatbutton,
				MAX_N_COUNTERS+3,MAX_N_COUNTERS+3,0,1,AlignHCenter );
	connect( repeatbutton, SIGNAL( clicked() ),
				this, SLOT( correct() ) );

	stopbutton = new QJKPushButton( "&Break",this );
	CHECK_PTR( stopbutton );
	stopbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addMultiCellWidget( stopbutton,
				MAX_N_COUNTERS+3,MAX_N_COUNTERS+3,2,3,AlignHCenter );
	connect( stopbutton, SIGNAL( clicked() ),
				&devmanager, SLOT( stopCounting() ) );

	acceptbutton = new QJKPushButton( "&Accept",this );
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addMultiCellWidget( acceptbutton,
				MAX_N_COUNTERS+4,MAX_N_COUNTERS+4,0,1,AlignHCenter );
	connect( acceptbutton, SIGNAL( clicked() ),
				this, SLOT( acceptCorrection() ) );

	cancelbutton = new QJKPushButton( "&Cancel",this );
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addMultiCellWidget( cancelbutton,
				MAX_N_COUNTERS+4,MAX_N_COUNTERS+4,2,3,AlignHCenter );	
	connect( cancelbutton, SIGNAL( clicked() ),
				this, SLOT( reject() ) );
	setButtonsEnabled( true );
}

void BackCorrectionDialog::update( void ) {
	int i;
	QString _str;
	devmanager.getBackgroundCorrectionCoefficients( &coef );
	for( i=0; i<MAX_N_COUNTERS; i++ ) {
		_str.setNum( coef[i][2],'f',2 );
		edit_l[i]->clear();
		edit_r[i]->clear();
		edit_av[i]->setText( _str );
	}
	timer = startTimer( 500 ); // 1 sec
}

void BackCorrectionDialog::finish( void ) {
	killTimer( timer );
}

void BackCorrectionDialog::acceptCorrection( void ) {
	devmanager.acceptBackgroundCorrection( &coef );
	accept();
}

void BackCorrectionDialog::timerEvent( QTimerEvent *e ) {
	stateExportMutex.lock();
	if( StateExport::isCorrectionReady ) {
		showResult();
		StateExport::isCorrectionReady = false;
	}
	stateExportMutex.unlock();
}

void BackCorrectionDialog::setButtonsEnabled( bool _enable ) {
	acceptbutton->setEnabled( _enable );
	cancelbutton->setEnabled( _enable );
	repeatbutton->setEnabled( _enable );
	stopbutton->setEnabled( !_enable );
	if( !_enable ) stopbutton->setFocus();
}

void BackCorrectionDialog::correct( void ) {
	debug( "slot repeat!" );
	devmanager.backgroundCorrection( &coef );
}

void BackCorrectionDialog::showResult( void ) {
	debug( "Result:" );
	int i;
	QString str;
	if( coef ) {
		for( i=0; i<MAX_N_COUNTERS; i++ ) {
			edit_l[i]->setText( str.setNum( coef[i][0],'f',2 ) );
			edit_r[i]->setText( str.setNum( coef[i][1],'f',2 ) );
			edit_av[i]->setText( str.setNum( coef[i][2],'f',2 ) );
		}
	}
}

BackCorrectionDialog::~BackCorrectionDialog() {
	delete label;
	delete label_cf;
	delete label_l;
	delete label_r;
	delete label_av;
	delete label_a;
	delete label_b;
	delete label_c;
	delete label_d;
	int i;
	for( i=0; i<MAX_N_COUNTERS; i++ ) {
		delete edit_l[i];
		delete edit_r[i];
		delete edit_av[i];
	}
	delete layout;
	delete acceptbutton;
	delete repeatbutton;
	delete cancelbutton;
	delete stopbutton;
}

CenterCorrectionDialog::CenterCorrectionDialog( void )
			: QDialog( 0,"Centering Correction",true ) {
	setCaption( "Centering Correction" );
	int border = fontMetrics().height();
	layout = new QGridLayout( this,5,2,border );
	CHECK_PTR(layout);

	label = new QLabel( "Centering Correction",this );
	CHECK_PTR( label );
	layout->addMultiCellWidget( label,0,0,0,1,AlignHCenter );

	label_x = new QLabel( "DeltaX0:",this );
	CHECK_PTR( label_x );
	layout->addWidget( label_x,1,0,AlignHCenter );

	label_y = new QLabel( "DeltaY0:",this );
	CHECK_PTR( label_y );
	layout->addWidget( label_y,2,0,AlignHCenter );

	int _width = fontMetrics().width( "wwww" );
	int _height = getHeight( fontMetrics().height() );

	edit_x = new QLineEdit( this );
	CHECK_PTR( edit_x );
	edit_x->setReadOnly( true );
	edit_x->setAlignment( AlignHCenter );
	edit_x->setFixedSize( _width, _height );
	edit_x->setFocusPolicy( QWidget::NoFocus );
	layout->addWidget( edit_x,1,1 );

	edit_y = new QLineEdit( this );
	CHECK_PTR( edit_y );
	edit_y->setReadOnly( true );
	edit_y->setAlignment( AlignHCenter );
	edit_y->setFixedSize( _width, _height );
	edit_y->setFocusPolicy( QWidget::NoFocus );
	layout->addWidget( edit_y,2,1 );

	repeatbutton = new QJKPushButton( "&Measure",this );
	CHECK_PTR(repeatbutton);
	repeatbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addWidget( repeatbutton,3,0 );
	connect( repeatbutton, SIGNAL( clicked() ),
				this, SLOT( correct() ) );

	stopbutton = new QJKPushButton( "&Break",this );
	CHECK_PTR( stopbutton );
	stopbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addWidget( stopbutton,3,1 );
	connect( stopbutton, SIGNAL( clicked() ),
				&devmanager, SLOT( stopCounting() ) );

	acceptbutton = new QJKPushButton( "&Accept",this );
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addWidget( acceptbutton,4,0 );
	connect( acceptbutton, SIGNAL( clicked() ),
				this, SLOT( acceptCorrection() ) );

	cancelbutton = new QJKPushButton( "&Cancel",this );
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize( repeatbutton->sizeHint() );
	layout->addWidget( cancelbutton,4,1 );
	connect( cancelbutton, SIGNAL( clicked() ),
				this, SLOT( reject() ) );
//	connect( &devmanager, SIGNAL( signalSendCorrection() ),
//				this, SLOT( showResult() ) );
	setButtonsEnabled( true );
}

void CenterCorrectionDialog::update( void ) {
	QString str;
	devmanager.getCenterCorrectionCoefficients( &delta );
	edit_x->setText( str.setNum( delta.x() ) );
	edit_y->setText( str.setNum( delta.y() ) );
	timer = startTimer( 500 ); // 1 sec
}

void CenterCorrectionDialog::finish( void ) {
	killTimer( timer );
}

void CenterCorrectionDialog::acceptCorrection( void ) {
	devmanager.acceptCenterCorrection( &delta );
	accept();
}

void CenterCorrectionDialog::timerEvent( QTimerEvent *e ) {
	stateExportMutex.lock();
	if( StateExport::isCorrectionReady ) {
		showResult();
		StateExport::isCorrectionReady = false;
	}
	stateExportMutex.unlock();
}

void CenterCorrectionDialog::setButtonsEnabled( bool _enable ) {
	acceptbutton->setEnabled( _enable );
	cancelbutton->setEnabled( _enable );
	repeatbutton->setEnabled( _enable );
	stopbutton->setEnabled( !_enable );
	if( !_enable ) stopbutton->setFocus();
}

void CenterCorrectionDialog::correct( void ) {
	debug( "slot repeat!" );
	devmanager.centerCorrection( &delta );
}

void CenterCorrectionDialog::showResult( void ) {
	debug( "Result:" );
	QString str;
	edit_x->setText( str.setNum( delta.x() ) );
	edit_y->setText( str.setNum( delta.y() ) );
}

CenterCorrectionDialog::~CenterCorrectionDialog() {
	delete label;
	delete label_x;
	delete label_y;
	delete edit_x;
	delete edit_y;
	delete layout;
	delete acceptbutton;
	delete repeatbutton;
	delete cancelbutton;
	delete stopbutton;
}
